# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
import warnings
from scipy.linalg import logm


def check_eos_consistency(m, P=1.e9, T=300., tol=1.e-4, verbose=False,
                          including_shear_properties=True):
    """
    Compute numerical derivatives of the gibbs free energy of a mineral
    under given conditions, and check these values against those provided
    analytically by the equation of state

    Parameters
    ----------
    m : mineral
        The mineral for which the equation of state
        is to be checked for consistency
    P : float
        The pressure at which to check consistency
    T : float
        The temperature at which to check consistency
    tol : float
        The fractional tolerance for each of the checks
    verbose : boolean
        Decide whether to print information about each
        check
    including_shear_properties : boolean
        Decide whether to check shear information,
        which is pointless for liquids and equations of state
        without shear modulus parameterizations

    Returns
    -------
    consistency: boolean
        If all checks pass, returns True

    """
    dT = 1.
    dP = 1000.

    m.set_state(P, T)
    G0 = m.gibbs
    S0 = m.S
    V0 = m.V

    expr = ['G = F + PV', 'G = H - TS', 'G = E - TS + PV']
    eq = [[m.gibbs, (m.helmholtz + P*m.V)],
          [m.gibbs, (m.H - T*m.S)],
          [m.gibbs, (m.molar_internal_energy - T*m.S + P*m.V)]]

    m.set_state(P, T + dT)
    G1 = m.gibbs
    S1 = m.S
    V1 = m.V

    m.set_state(P + dP, T)
    G2 = m.gibbs
    V2 = m.V

    # T derivatives
    m.set_state(P, T + 0.5*dT)
    expr.extend(['S = -dG/dT', 'alpha = 1/V dV/dT', 'C_p = T dS/dT'])
    eq.extend([[m.S, -(G1 - G0)/dT],
               [m.alpha, (V1 - V0)/dT/m.V],
               [m.molar_heat_capacity_p, (T + 0.5*dT)*(S1 - S0)/dT]])

    # P derivatives
    m.set_state(P + 0.5*dP, T)
    expr.extend(['V = dG/dP', 'K_T = -V dP/dV'])
    eq.extend([[m.V, (G2 - G0)/dP],
               [m.K_T, -0.5*(V2 + V0)*dP/(V2 - V0)]])

    expr.extend(['C_v = Cp - alpha^2*K_T*V*T', 'K_S = K_T*Cp/Cv',
                 'gr = alpha*K_T*V/Cv'])
    eq.extend([[m.molar_heat_capacity_v,
                m.molar_heat_capacity_p - m.alpha*m.alpha*m.K_T*m.V*T],
               [m.K_S, m.K_T*m.molar_heat_capacity_p/m.molar_heat_capacity_v],
               [m.gr, m.alpha*m.K_T*m.V/m.molar_heat_capacity_v]])

    expr.append('Vphi = np.sqrt(K_S/rho)')
    eq.append([m.bulk_sound_velocity, np.sqrt(m.K_S/m.rho)])

    if including_shear_properties:
        expr.extend(['Vp = np.sqrt((K_S + 4G/3)/rho)',
                     'Vs = np.sqrt(G_S/rho)'])

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            eq.extend([[m.p_wave_velocity, np.sqrt((m.K_S + 4.*m.G/3.)/m.rho)],
                       [m.shear_wave_velocity, np.sqrt(m.G/m.rho)]])
            if len(w) == 1:
                print(w[0].message)
                print('\nYou can suppress this message by setting the '
                      'parameter\nincluding_shear_properties to False '
                      'when calling check_eos_consistency.\n')
        note = ''
    else:
        note = ' (not including shear properties)'

    consistencies = [np.abs(e[0] - e[1])
                     < np.abs(tol*e[1]) + np.finfo('float').eps for e in eq]
    eos_is_consistent = np.all(consistencies)

    if verbose:
        print('Checking EoS consistency for {0:s}{1}'.format(m.to_string(),
                                                             note))
        print('Expressions within tolerance of {0:2f}'.format(tol))
        for i, c in enumerate(consistencies):
            print('{0:10s} : {1:5s}'.format(expr[i], str(c)))
        if eos_is_consistent:
            print('All EoS consistency constraints satisfied for {0:s}'.format(m.to_string()))
        else:
            print('Not satisfied all EoS consistency constraints for {0:s}'.format(m.to_string()))

    return eos_is_consistent


def check_anisotropic_eos_consistency(m, P=1.e9, T=2000.,
                                      tol=1.e-4, verbose=False):
    """
    Compute numerical derivatives of the gibbs free energy of a mineral
    under given conditions, and check these values against those provided
    analytically by the equation of state

    Parameters
    ----------
    m : mineral
        The mineral for which the equation of state
        is to be checked for consistency
    P : float
        The pressure at which to check consistency
    T : float
        The temperature at which to check consistency
    tol : float
        The fractional tolerance for each of the checks
    verbose : boolean
        Decide whether to print information about each
        check

    Returns
    -------
    consistency: boolean
        If all checks pass, returns True

    """
    dT = 1.
    dP = 1000.

    m.set_state(P, T)
    G0 = m.gibbs
    S0 = m.S
    V0 = m.V

    expr = ['G = F + PV', 'G = H - TS', 'G = E - TS + PV']
    eq = [[m.gibbs, (m.helmholtz + P*m.V)],
          [m.gibbs, (m.H - T*m.S)],
          [m.gibbs, (m.molar_internal_energy - T*m.S + P*m.V)]]

    m.set_state(P, T + dT)
    G1 = m.gibbs
    S1 = m.S
    V1 = m.V

    m.set_state(P + dP, T)
    G2 = m.gibbs
    V2 = m.V

    # T derivatives
    m.set_state(P, T + 0.5*dT)
    expr.extend(['S = -dG/dT', 'alpha = 1/V dV/dT', 'C_p = T dS/dT'])
    eq.extend([[m.S, -(G1 - G0)/dT],
               [m.alpha, (V1 - V0)/dT/m.V],
               [m.molar_heat_capacity_p, (T + 0.5*dT)*(S1 - S0)/dT]])

    # P derivatives
    m.set_state(P + 0.5*dP, T)
    expr.extend(['V = dG/dP', 'K_T = -V dP/dV'])
    eq.extend([[m.V, (G2 - G0)/dP],
               [m.isothermal_bulk_modulus_reuss, -0.5*(V2 + V0)*dP/(V2 - V0)]])

    expr.extend(['C_v = Cp - alpha^2*K_T*V*T', 'K_S = K_T*Cp/Cv'])
    eq.extend([[m.molar_heat_capacity_v,
                m.molar_heat_capacity_p - m.alpha*m.alpha*m.K_T*m.V*T],
               [m.isentropic_bulk_modulus_reuss,
                m.isothermal_bulk_modulus_reuss
                * m.molar_heat_capacity_p / m.molar_heat_capacity_v]])

    # Third derivative
    m.set_state(P + 0.5 * dP, T)
    b0 = m.isothermal_compressibility_tensor
    F0 = m.deformation_gradient_tensor

    m.set_state(P + 0.5 * dP, T + dT)
    b1 = m.isothermal_compressibility_tensor
    F1 = m.deformation_gradient_tensor

    m.set_state(P, T + 0.5 * dT)
    a0 = m.thermal_expansivity_tensor
    F2 = m.deformation_gradient_tensor

    m.set_state(P + dP, T + 0.5 * dT)
    a1 = m.thermal_expansivity_tensor
    F3 = m.deformation_gradient_tensor

    m.set_state(P + 0.5 * dP, T + 0.5 * dT)

    beta0 = -(logm(F3) - logm(F2))/dP
    alpha0 = (logm(F1) - logm(F0))/dT

    Q = m.deformed_coordinate_frame
    beta1 = m.isothermal_compressibility_tensor
    alpha1 = m.thermal_expansivity_tensor

    beta1 = np.einsum('mi, nj, ij->mn', Q, Q, beta1)
    alpha1 = np.einsum('mi, nj, ij->mn', Q, Q, alpha1)

    expr.extend([f'SI = -d(lnm(F))/dP ({i}{j})'
                 for i in range(3) for j in range(i, 3)])
    eq.extend([[beta0[i, j], beta1[i, j]]
               for i in range(3) for j in range(i, 3)])

    expr.extend([f'alpha = d(lnm(F))/dT ({i}{j})'
                 for i in range(3) for j in range(i, 3)])
    eq.extend([[alpha0[i, j], alpha1[i, j]]
               for i in range(3) for j in range(i, 3)])

    expr.extend([f'd(alpha)/dP = -d(beta_T)/dT ({i}{j})'
                 for i in range(3) for j in range(i, 3)])
    eq.extend([[(a1[i, j] - a0[i, j])/dP, -(b1[i, j] - b0[i, j])/dT]
               for i in range(3) for j in range(i, 3)])

    # Consistent isotropic and anisotropic properties
    expr.extend(['V = det(M)',
                 'alpha_v = tr(alpha)',
                 'beta_T = sum(S_T I)',
                 'beta_S = sum(S_S I)'])
    eq.extend([[m.V, np.linalg.det(m.cell_vectors)],
               [m.alpha, np.trace(m.thermal_expansivity_tensor)],
               [m.beta_T, np.sum(m.isothermal_compliance_tensor[:3, :3])],
               [m.beta_S, np.sum(m.isentropic_compliance_tensor[:3, :3])]])

    expr.append('Vphi = np.sqrt(K_S/rho)')
    eq.append([m.bulk_sound_velocity, np.sqrt(m.K_S/m.rho)])

    consistencies = [np.abs(e[0] - e[1]) < np.abs(tol*e[1])
                     + np.finfo('float').eps for e in eq]
    eos_is_consistent = np.all(consistencies)

    if verbose:
        print('Checking EoS consistency for {0:s}'.format(m.to_string()))
        print('Expressions within tolerance of {0:2f}'.format(tol))
        for i, c in enumerate(consistencies):
            print('{0:10s} : {1:5s}'.format(expr[i], str(c)))
        if eos_is_consistent:
            print('All EoS consistency constraints satisfied for {0:s}'.format(m.to_string()))
        else:
            print('Not satisfied all EoS consistency constraints for {0:s}'.format(m.to_string()))

    return eos_is_consistent



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.
import numpy as np
from scipy.linalg import expm
from numpy.linalg import cond
from .mineral import Mineral
from .material import Material, material_property
from .anisotropy import AnisotropicMaterial, voigt_compliance_factors
from ..tools.misc import copy_documentation
from ..tools.unitcell import cell_parameters_to_vectors
from ..tools.unitcell import cell_vectors_to_parameters

class AnisotropicMineral(Mineral, AnisotropicMaterial):
    """
    A class implementing the anisotropic mineral equation of state described
    in :cite:`Myhill2021`.
    This class is derived from both Mineral and AnisotropicMaterial,
    and inherits most of the methods from these classes.

    Instantiation of an AnisotropicMineral takes three arguments;
    a reference Mineral (i.e. a standard isotropic mineral which provides
    volume as a function of pressure and temperature), cell_parameters,
    which give the lengths of the molar cell vectors and the angles between
    them (see :func:`~burnman.tools.unitcell.cell_parameters_to_vectors`),
    and a 4D array of anisotropic parameters which describe the
    anisotropic behaviour of the mineral. For a description of the physical
    meaning of these parameters, please refer to the code
    or to the original paper.

    States of the mineral can only be queried after setting the
    pressure and temperature using set_state().

    This class is available as ``burnman.AnisotropicMineral``.

    All the material parameters are expected to be in plain SI units.  This
    means that the elastic moduli should be in Pascals and NOT Gigapascals.
    Additionally, the cell parameters should be in m/(mol formula unit)
    and not in unit cell lengths. To convert unit cell lengths given in
    Angstrom to molar cell parameters you should multiply by 10^(-10) *
    (N_a / Z)^1/3, where N_a is Avogadro's number
    and Z is the number of formula units per unit cell.
    You can look up Z in many places, including www.mindat.org.

    Finally, it is assumed that the unit cell of the anisotropic material
    is aligned in a particular way relative to the coordinate axes
    (the anisotropic_parameters are defined relative to the coordinate axes).
    The crystallographic a-axis is assumed to be parallel to the first
    spatial coordinate axis, and the crystallographic b-axis is assumed to
    be perpendicular to the third spatial coordinate axis.
    """

    def __init__(self, isotropic_mineral, cell_parameters,
                 anisotropic_parameters):

        if not np.all(anisotropic_parameters[:, :, 0, 0] == 0):
            raise Exception("anisotropic_parameters_pqmn should be set to "
                            "zero for all m = n = 0")

        sum_ijij_block = np.sum(anisotropic_parameters[:3, :3, :, :],
                                axis=(0, 1))

        if np.abs(sum_ijij_block[1, 0] - 1.) > 1.e-5:
            raise Exception('The sum of the upper 3x3 pq-block of '
                            'anisotropic_parameters_pqmn must equal '
                            '1 for m=1, n=0 for consistency with the volume. '
                            f'Value is {sum_ijij_block[1, 0]}')

        for m in range(2, len(sum_ijij_block)):
            if np.abs(sum_ijij_block[m, 0]) > 1.e-10:
                raise Exception('The sum of the upper 3x3 pq-block of '
                                'anisotropic_parameters_pqmn must equal 0 for'
                                f'm={m}, n=0 for consistency with the volume. '
                                f'Value is {sum_ijij_block[m, 0]}')

        for m in range(len(sum_ijij_block)):
            for n in range(1, len(sum_ijij_block[0])):
                if np.abs(sum_ijij_block[m, n]) > 1.e-10:
                    raise Exception('The sum of the upper 3x3 pq-block of '
                                    'anisotropic_parameters_pqmn must equal '
                                    f'0 for m={m}, n={n} for '
                                    'consistency with the volume. '
                                    f'Value is {sum_ijij_block[m, n]}')

        if cond(anisotropic_parameters[:, :, 1, 0]) > 1/np.finfo(float).eps:
            raise Exception('anisotropic_parameters[:, :, 1, 0] is singular')

        sum_lower_left_block = np.sum(anisotropic_parameters[3:, :3, :, :],
                                      axis=1)

        self.orthotropic = True
        for i, s in enumerate(sum_lower_left_block):
            if not np.all(np.abs(s) < 1.e-10):
                self.orthotropic = False

        self.cell_vectors_0 = cell_parameters_to_vectors(cell_parameters)

        if (np.abs(np.linalg.det(self.cell_vectors_0)
                   - isotropic_mineral.params['V_0']) > np.finfo(float).eps):
            factor = np.cbrt(isotropic_mineral.params['V_0']
                             / np.linalg.det(self.cell_vectors_0))
            raise Exception('The standard state unit vectors are inconsistent '
                            'with the volume. Suggest multiplying each '
                            f'by {factor}.')

        self.c = anisotropic_parameters

        if 'name' in isotropic_mineral.params:
            self.name = isotropic_mineral.params['name']

        Mineral.__init__(self, isotropic_mineral.params,
                         isotropic_mineral.property_modifiers)

    @copy_documentation(Material.set_state)
    def set_state(self, pressure, temperature):
        # 1) Compute dPthdf|T
        # relatively large dP needed for accurate estimate of dPthdf
        Mineral.set_state(self, pressure, temperature)
        dP = self.isothermal_bulk_modulus_reuss*1.e-5

        Mineral.set_state(self, pressure-dP/2., temperature)
        V1 = self.V
        Pth1 = pressure-dP/2. - self.method.pressure(self.params['T_0'],
                                                     V1, self.params)

        Mineral.set_state(self, pressure+dP/2., temperature)
        V2 = self.V
        Pth2 = pressure+dP/2. - self.method.pressure(self.params['T_0'],
                                                     V2, self.params)

        self.dPthdf = (Pth2 - Pth1) / np.log(V2/V1)
        Mineral.set_state(self, pressure, temperature)

        # 2) Compute other properties needed for anisotropic equation of state
        V = self.V
        V_0 = self.params['V_0']
        Vrel = V/V_0
        f = np.log(Vrel)

        Pth = pressure - self.method.pressure(self.params['T_0'], self.V,
                                              self.params)

        # Compute X, dXdPth, dXdf, needed by most anisotropic properties
        ns = np.arange(self.c.shape[-1])
        x = self.c[:, :, 0, :] + self.c[:, :, 1, :]*f
        dxdf = self.c[:, :, 1, :]

        for i in list(range(2, self.c.shape[2])):
            # non-intuitively, the += operator doesn't simply add in-place,
            # so here we overwrite the arrays with new ones
            x = x + self.c[:, :, i, :]*np.power(f, float(i))/float(i)
            dxdf = dxdf + self.c[:, :, i, :] * np.power(f, float(i)-1.)

        self._Vrel = Vrel
        self._f = f
        self.X_Voigt = np.einsum('ikn, n->ik', x, np.power(Pth, ns))
        self.dXdPth_Voigt = np.einsum('ikn, n->ik',
                                      x[:, :, 1:],
                                      ns[1:]*np.power(Pth, ns[1:]-1))
        self.dXdf_Voigt = np.einsum('ikn, n->ik', dxdf, np.power(Pth, ns))

    def _contract_compliances(self, compliances):
        """
        Takes a compliance tensor in standard (3x3x3x3) form
        and returns the Voigt form (6x6). Note the compliance factors
        which are required to maintain the inverse relationship with the
        corresponding stiffness tensor.
        """
        voigt_notation = np.zeros((6, 6))
        for p in range(6):
            i, j = self._voigt_index_to_ij(p)
            for q in range(6):
                m, n = self._voigt_index_to_ij(q)
                voigt_notation[p, q] = compliances[i, j, m, n]
        return np.multiply(voigt_notation, voigt_compliance_factors)

    @material_property
    def deformation_gradient_tensor(self):
        """
        Returns
        -------
        deformation_gradient_tensor : 2D numpy array
            The deformation gradient tensor describing the deformation of the
            mineral from its undeformed state
            (i.e. the state at the reference pressure and temperature).
        """
        X_full = self._voigt_notation_to_compliance_tensor(self.X_Voigt)
        F = expm(np.einsum('ijkl, kl', X_full, np.eye(3)))
        return F

    @material_property
    def unrotated_cell_vectors(self):
        """
        Returns
        -------
        unrotated_cell_vectors : 2D numpy array
            The vectors of the cell constructed from one mole of formula units
            after deformation of the mineral from its undeformed state
            (i.e. the state at the reference pressure and temperature).
            Each vector is given in [m]. See the documentation for the function
            :func:`~burnman.tools.unitcell.cell_parameters_to_vectors`
            for the assumed relationships between the cell vectors and
            spatial coordinate axes.
        """
        return self.deformation_gradient_tensor.dot(self.cell_vectors_0)

    @material_property
    def deformed_coordinate_frame(self):
        """
        Returns
        -------
        deformed_coordinate_frame : 2D numpy array
            The orientations of the three spatial coordinate axes
            after deformation of the mineral. For orthotropic minerals,
            this is equal to the identity matrix, as hydrostatic stresses only
            induce rotations in monoclinic and triclinic crystals.
        """
        if self.orthotropic:
            return np.eye(3)
        else:
            M = self.unrotated_cell_vectors
            Q = np.empty((3, 3))
            Q[0] = M[0] / np.linalg.norm(M[0])
            Q[2] = np.cross(M[0], M[1])/np.linalg.norm(np.cross(M[0], M[1]))
            Q[1] = np.cross(Q[2], Q[0])
            return Q

    @material_property
    def rotation_matrix(self):
        """
        Returns
        -------
        rotation_matrix : 2D numpy array
            The matrix required to rotate the properties of the deformed
            mineral into the deformed coordinate frame. For orthotropic
            minerals, this is equal to the identity matrix.
        """
        return self.deformed_coordinate_frame.T

    @material_property
    def cell_vectors(self):
        """
        Returns
        -------
        cell_vectors : 2D numpy array
            The vectors of the cell constructed from one mole of formula units.
            Each vector is given in [m]. See the documentation for the function
            :func:`~burnman.tools.unitcell.cell_parameters_to_vectors`
            for the assumed relationships between the cell vectors and
            spatial coordinate axes.
        """
        if self.orthotropic:
            return self.unrotated_cell_vectors
        else:
            return np.einsum('ij, jk->ik',
                             self.unrotated_cell_vectors,
                             self.rotation_matrix)

    @material_property
    def cell_parameters(self):
        """
        Returns
        -------
        cell_parameters : 1D numpy array
            The molar cell parameters of the mineral, given in standard form:
            [:math:`a`, :math:`b`, :math:`c`,
            :math:`\\alpha`, :math:`\\beta`, :math:`\\gamma`],
            where the first three floats are the lengths of the vectors in [m]
            defining the cell constructed from one mole of formula units.
            The last three floats are angles between vectors
            (given in radians). See the documentation for the function
            :func:`~burnman.tools.unitcell.cell_parameters_to_vectors`
            for the assumed relationships between the cell vectors and
            spatial coordinate axes.
        """
        return cell_vectors_to_parameters(self.cell_vectors)

    @material_property
    def shear_modulus(self):
        """
        Anisotropic minerals do not (in general) have a single shear modulus.
        This function returns a NotImplementedError. Users should instead
        consider directly querying the elements in the
        isothermal_stiffness_tensor or isentropic_stiffness_tensor.
        """
        raise NotImplementedError("Anisotropic minerals do not have a shear "
                                  "modulus property. Query "
                                  "the isentropic or isothermal stiffness "
                                  "tensors directory, or use"
                                  "isentropic_shear_modulus_reuss or "
                                  "isentropic_shear_modulus_voigt.")

    @material_property
    def isothermal_bulk_modulus(self):
        """
        Anisotropic minerals do not have a single isothermal bulk modulus.
        This function returns a NotImplementedError. Users should instead
        consider either using isothermal_bulk_modulus_reuss,
        isothermal_bulk_modulus_voigt,
        or directly querying the elements in the isothermal_stiffness_tensor.
        """
        raise NotImplementedError("isothermal_bulk_modulus is not "
                                  "sufficiently explicit for an "
                                  "anisotropic mineral. Did you mean "
                                  "isothermal_bulk_modulus_reuss?")

    @material_property
    def isentropic_bulk_modulus(self):
        """
        Anisotropic minerals do not have a single isentropic bulk modulus.
        This function returns a NotImplementedError. Users should instead
        consider either using isentropic_bulk_modulus_reuss,
        isentropic_bulk_modulus_voigt (both derived from Anisotropicmineral),
        or directly querying the elements in the isentropic_stiffness_tensor.
        """
        raise NotImplementedError("isentropic_bulk_modulus is not "
                                  "sufficiently explicit for an "
                                  "anisotropic mineral. Did you mean "
                                  "isentropic_bulk_modulus_reuss?")

    isothermal_bulk_modulus_reuss = Mineral.isothermal_bulk_modulus

    @material_property
    def isothermal_bulk_modulus_voigt(self):
        """
        Returns
        -------
        isothermal_bulk_modulus_voigt : float
            The Voigt bound on the isothermal bulk modulus in [Pa].
        """
        K = np.sum([[self.isothermal_stiffness_tensor[i][k]
                     for k in range(3)]
                    for i in range(3)])/9.
        return K

    @material_property
    def isothermal_compressibility_reuss(self):
        """
        Returns
        -------
        isothermal_compressibility_reuss : float
            The Reuss bound on the isothermal compressibility in [1/Pa].
        """
        return 1./self.isothermal_bulk_modulus_reuss

    @material_property
    def isothermal_compliance_tensor(self):
        """
        Returns
        -------
        isothermal_compliance_tensor : 2D numpy array
            The isothermal compliance tensor [1/Pa]
            in Voigt form (:math:`\\mathbb{S}_{\\text{T} pq}`).
        """
        S_T = (self.isothermal_compressibility_reuss
               * (self.dXdf_Voigt + self.dXdPth_Voigt * self.dPthdf))
        if self.orthotropic:
            return S_T
        else:
            R = self.rotation_matrix
            S = self._voigt_notation_to_compliance_tensor(S_T)
            S_rotated = np.einsum('mi, nj, ok, pl, ijkl->mnop', R, R, R, R, S)
            return self._contract_compliances(S_rotated)

    @material_property
    def thermal_expansivity_tensor(self):
        """
        Returns
        -------
        thermal_expansivity_tensor : 2D numpy array
            The tensor of thermal expansivities [1/K].
        """
        a = self.alpha * (self.dXdf_Voigt
                          + self.dXdPth_Voigt
                          * (self.dPthdf
                             + 1./self.isothermal_compressibility_reuss))
        alpha = np.einsum('ijkl, kl',
                          self._voigt_notation_to_compliance_tensor(a),
                          np.eye(3))

        if self.orthotropic:
            return alpha
        else:
            R = self.rotation_matrix
            return np.einsum('mi, nj, ij->mn', R, R, alpha)

    # Derived properties start here
    @material_property
    def isothermal_stiffness_tensor(self):
        """
        Returns
        -------
        isothermak_stiffness_tensor : 2D numpy array
            The isothermal stiffness tensor [Pa]
            in Voigt form (:math:`\\mathbb{C}_{\\text{T} pq}`).
        """
        return np.linalg.inv(self.isothermal_compliance_tensor)

    @material_property
    def full_isothermal_compliance_tensor(self):
        """
        Returns
        -------
        full_isothermak_stiffness_tensor : 4D numpy array
            The isothermal compliance tensor [1/Pa]
            in standard form (:math:`\\mathbb{S}_{\\text{T} ijkl}`).
        """
        S_Voigt = self.isothermal_compliance_tensor
        return self._voigt_notation_to_compliance_tensor(S_Voigt)

    @material_property
    def full_isothermal_stiffness_tensor(self):
        """
        Returns
        -------
        full_isothermak_stiffness_tensor : 4D numpy array
            The isothermal stiffness tensor [Pa]
            in standard form (:math:`\\mathbb{C}_{\\text{T} ijkl}`).
        """
        return self._expand_stiffnesses(self.isothermal_stiffness_tensor)

    @material_property
    def full_isentropic_compliance_tensor(self):
        """
        Returns
        -------
        full_isothermak_stiffness_tensor : 4D numpy array
            The isentropic compliance tensor [1/Pa]
            in standard form (:math:`\\mathbb{S}_{\\text{N} ijkl}`).
        """
        return (self.full_isothermal_compliance_tensor
                - np.einsum('ij, kl->ijkl',
                            self.thermal_expansivity_tensor,
                            self.thermal_expansivity_tensor)
                * self.V * self.temperature / self.C_p)

    @material_property
    def isentropic_compliance_tensor(self):
        """
        Returns
        -------
        isentropic_compliance_tensor : 2D numpy array
            The isentropic compliance tensor [1/Pa]
            in Voigt form (:math:`\\mathbb{S}_{\\text{N} pq}`).
        """
        S_full = self.full_isentropic_compliance_tensor
        return self._contract_compliances(S_full)

    @material_property
    def isentropic_stiffness_tensor(self):
        """
        Returns
        -------
        isentropic_stiffness_tensor : 2D numpy array
            The isentropic stiffness tensor [Pa]
            in Voigt form (:math:`\\mathbb{C}_{\\text{N} pq}`).
        """
        return np.linalg.inv(self.isentropic_compliance_tensor)

    @material_property
    def full_isentropic_stiffness_tensor(self):
        """
        Returns
        -------
        full_isentropic_stiffness_tensor : 4D numpy array
            The isentropic stiffness tensor [Pa]
            in standard form (:math:`\\mathbb{C}_{\\text{N} ijkl}`).
        """
        C_Voigt = self.isentropic_stiffness_tensor
        return self._voigt_notation_to_stiffness_tensor(C_Voigt)

    @material_property
    def grueneisen_tensor(self):
        """
        Returns
        -------
        grueneisen_tensor : 2D numpy array
            The grueneisen tensor.
            This is defined by :cite:`BarronMunn1967` as
            :math:`\\mathbb{C}_{\\text{N} ijkl} \\alpha_{kl} V/C_{P}`.
        """
        return (np.einsum('ijkl, kl->ij',
                          self.full_isentropic_stiffness_tensor,
                          self.thermal_expansivity_tensor)
                * self.molar_volume / self.molar_heat_capacity_p)

    @material_property
    def grueneisen_parameter(self):
        """
        Anisotropic minerals do not (in general) have a single grueneisen
        parameter. This function returns a NotImplementedError.
        Users should instead consider directly querying the elements in the
        grueneisen_tensor.
        """
        raise NotImplementedError("Anisotropic minerals do not have a single "
                                  "grueneisen parameter. Query "
                                  "the grueneisen_tensor instead.")

    @material_property
    def isothermal_compressibility_tensor(self):
        """
        Returns
        -------
        isothermal_compressibility_tensor : 2D numpy array
            The isothermal compressibility tensor.
        """
        return np.einsum('ijkl, kl->ij',
                         self.full_isothermal_compliance_tensor,
                         np.eye(3))

    @material_property
    def isentropic_compressibility_tensor(self):
        """
        Returns
        -------
        isentropic_compressibility_tensor : 2D numpy array
            The isentropic compressibility tensor.
        """
        return np.einsum('ijkl, kl->ij',
                         self.full_isentropic_compliance_tensor,
                         np.eye(3))



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


class EquationOfState(object):

    """
    This class defines the interface for an equation of state
    that a mineral uses to determine its properties at a
    given :math:`P, T`.  In order define a new equation of state, you
    should define these functions.

    All functions should accept and return values in SI units.

    In general these functions are functions of pressure,
    temperature, and volume, as well as a "params" object,
    which is a Python dictionary that stores the material
    parameters of the mineral, such as reference volume,
    Debye temperature, reference moduli, etc.

    The functions for volume and density are just functions
    of temperature, pressure, and "params"; after all, it
    does not make sense for them to be functions of volume or density.
    """

    def volume(self, pressure, temperature, params):
        """
        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the equation of state. :math:`[Pa]`
        temperature : float
            Temperature at which to evaluate the equation of state. :math:`[K]`
        params : dictionary
            Dictionary containing material parameters required by the equation of state.

        Returns
        -------
        volume : float
            Molar volume of the mineral. :math:`[m^3]`
        """
        raise NotImplementedError("")

    def pressure(self, temperature, volume, params):
        """
        Parameters
        ----------
        volume : float
            Molar volume at which to evaluate the equation of state. [m^3]
        temperature : float
            Temperature at which to evaluate the equation of state. [K]
        params : dictionary
            Dictionary containing material parameters required by the equation of state.

        Returns
        -------
        pressure : float
            Pressure of the mineral, including cold and thermal parts. [m^3]
        """
        raise NotImplementedError("")

    def density(self, volume, params):
        """
        Calculate the density of the mineral :math:`[kg/m^3]`.
        The params object must include a "molar_mass" field.

        Parameters
        ----------
        volume : float
        Molar volume of the mineral.  For consistency this should be calculated
        using :func:`volume`. :math:`[m^3]`
        params : dictionary
            Dictionary containing material parameters required by the equation of state.

        Returns
        -------
        density : float
            Density of the mineral. :math:`[kg/m^3]`
        """
        return params["molar_mass"] / volume

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the equation of state. :math:`[Pa]`
        temperature : float
            Temperature at which to evaluate the equation of state. :math:`[K]`
        volume : float
            Molar volume of the mineral.  For consistency this should be calculated
            using :func:`volume`. :math:`[m^3]`
        params : dictionary
            Dictionary containing material parameters required by the equation of state.

        Returns
        -------
        gamma : float
            Grueneisen parameter of the mineral. :math:`[unitless]`
        """
        raise NotImplementedError("")

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the equation of state. :math:`[Pa]`
        temperature : float
            Temperature at which to evaluate the equation of state. :math:`[K]`
        volume : float
            Molar volume of the mineral.  For consistency this should be calculated
            using :func:`volume`. :math:`[m^3]`
        params : dictionary
            Dictionary containing material parameters required by the equation of state.

        Returns
        -------
        K_T : float
            Isothermal bulk modulus of the mineral. :math:`[Pa]`
        """
        raise NotImplementedError("")

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the equation of state. :math:`[Pa]`
        temperature : float
            Temperature at which to evaluate the equation of state. :math:`[K]`
        volume : float
            Molar volume of the mineral.  For consistency this should be calculated
            using :func:`volume`. :math:`[m^3]`
        params : dictionary
            Dictionary containing material parameters required by the equation of state.

        Returns
        -------
        K_S : float
            Adiabatic bulk modulus of the mineral. :math:`[Pa]`
        """
        raise NotImplementedError("")

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the equation of state. :math:`[Pa]`
        temperature : float
            Temperature at which to evaluate the equation of state. :math:`[K]`
        volume : float
            Molar volume of the mineral.  For consistency this should be calculated
            using :func:`volume`. :math:`[m^3]`
        params : dictionary
            Dictionary containing material parameters required by the equation of state.

        Returns
        -------
        G : float
            Shear modulus of the mineral. :math:`[Pa]`
        """
        raise NotImplementedError("")

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the equation of state. :math:`[Pa]`
        temperature : float
            Temperature at which to evaluate the equation of state. :math:`[K]`
        volume : float
            Molar volume of the mineral.  For consistency this should be calculated
            using :func:`volume`. :math:`[m^3]`
        params : dictionary
            Dictionary containing material parameters required by the equation of state.

        Returns
        -------
        C_V : float
            Heat capacity at constant volume of the mineral. :math:`[J/K/mol]`
        """
        raise NotImplementedError("")

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the equation of state. :math:`[Pa]`
        temperature : float
            Temperature at which to evaluate the equation of state. :math:`[K]`
        volume : float
            Molar volume of the mineral.  For consistency this should be calculated
            using :func:`volume`. :math:`[m^3]`
        params : dictionary
            Dictionary containing material parameters required by the equation of state.

        Returns
        -------
        C_P : float
            Heat capacity at constant pressure of the mineral. :math:`[J/K/mol]`
        """
        raise NotImplementedError("")

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the equation of state. :math:`[Pa]`
        temperature : float
            Temperature at which to evaluate the equation of state. :math:`[K]`
        volume : float
            Molar volume of the mineral.  For consistency this should be calculated
            using :func:`volume`. :math:`[m^3]`
        params : dictionary
            Dictionary containing material parameters required by the equation of state.

        Returns
        -------
        alpha : float
            Thermal expansivity of the mineral. :math:`[1/K]`
        """
        raise NotImplementedError("")

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the equation of state. [Pa]
        temperature : float
            Temperature at which to evaluate the equation of state. [K]
        volume : float
            Molar volume of the mineral.  For consistency this should be calculated
            using :func:`volume`. [m^3]
        params : dictionary
            Dictionary containing material parameters required by the equation of state.

        Returns
        -------
        G : float
            Gibbs free energy of the mineral
        """
        raise NotImplementedError("")

    def helmholtz_free_energy(self, pressure, temperature, volume, params):
        """
        Parameters
        ----------
        temperature : float
            Temperature at which to evaluate the equation of state. [K]
        volume : float
            Molar volume of the mineral.  For consistency this should be calculated
            using :func:`volume`. [m^3]
        params : dictionary
            Dictionary containing material parameters required by the equation of state.

        Returns
        -------
        F : float
            Helmholtz free energy of the mineral
        """
        raise NotImplementedError("")

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the entropy at the pressure and temperature of the mineral [J/K/mol]
        """

        raise NotImplementedError("")

    def enthalpy(self, pressure, temperature, volume, params):
        """
        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the equation of state. [Pa]
        temperature : float
            Temperature at which to evaluate the equation of state. [K]
        params : dictionary
            Dictionary containing material parameters required by the equation of state.

        Returns
        -------
        H : float
            Enthalpy of the mineral
        """
        raise NotImplementedError("")

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the equation of state. [Pa]
        temperature : float
            Temperature at which to evaluate the equation of state. [K]
        volume : float
            Molar volume of the mineral.  For consistency this should be calculated
            using :func:`volume`. [m^3]
        params : dictionary
            Dictionary containing material parameters required by the equation of state.

        Returns
        -------
        U : float
            Internal energy of the mineral
        """
        raise NotImplementedError("")

    def validate_parameters(self, params):
        """
        The params object is just a dictionary associating mineral physics parameters
        for the equation of state.  Different equation of states can have different parameters,
        and the parameters may have ranges of validity.  The intent of this function is
        twofold. First, it can check for the existence of the parameters that the
        equation of state needs, and second, it can check whether the parameters have reasonable
        values.  Unreasonable values will frequently be due to unit issues (e.g., supplying
        bulk moduli in GPa instead of Pa). In the base class this function does nothing,
        and an equation of state is not required to implement it.  This function will
        not return anything, though it may raise warnings or errors.

        Parameters
        ----------
        params : dictionary
            Dictionary containing material parameters required by the equation of state.
        """
        pass



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


from __future__ import absolute_import

import numpy as np
import warnings

from .mineral import Mineral, material_property
from .solidsolution import SolidSolution
from .solutionmodel import *
from .averaging_schemes import reuss_average_function

from ..tools.chemistry import sum_formulae
from .. import constants


class CombinedMineral(Mineral):

    """
    This is the base class for endmembers constructed from a
    linear combination of other minerals.

    Instances of this class should be initialised with a
    list of Mineral instances, a second list containing the
    number of moles of each mineral, and (optionally) a
    third list containing three floats describing a
    free energy adjustment which is linear in pressure and temperature
    (i.e. a constant energy [J/mol], entropy [J/K/mol]
    and volume adjustment [J/Pa/mol or m^3/mol]).

    For example, a crude approximation to a bridgmanite model might be
    bdg = CombinedMineral([per, stv], [1.0, 1.0], [-15.e3, 0., 0.])

    This class is available as :class:`burnman.CombinedMineral`.
    """

    def __init__(self, mineral_list, molar_amounts, free_energy_adjustment=[], name='User-created endmember'):
        self.mixture = SolidSolution(solution_type='mechanical',
                                     endmembers=[[m, ''] for m in mineral_list],
                                     molar_fractions = molar_amounts)

        # Remove elements from the chemical formula if they have negligible amounts
        for key, value in list(self.mixture.formula.items()):
            if np.abs(value) < 1.e-10:
                self.mixture.formula.pop(key)

        self.params = {
            'name': name,
            'formula': self.mixture.formula,
            'equation_of_state': 'combined',
            'molar_mass': self.mixture.molar_mass,
            'n': sum(self.mixture.formula.values())
        }

        if free_energy_adjustment != []:
            assert(len(free_energy_adjustment) == 3)
            self.property_modifiers = [['linear', {'delta_E': free_energy_adjustment[0],
                                                   'delta_S': free_energy_adjustment[1],
                                                   'delta_V': free_energy_adjustment[2]}]]


        Mineral.__init__(self)

    def set_state(self, pressure, temperature):
        self.mixture.set_state(pressure, temperature)
        Mineral.set_state(self, pressure, temperature)

    @material_property
    def molar_gibbs(self):
        """
        Returns Gibbs free energy of the solid solution [J]
        Aliased with self.gibbs
        """
        return self.mixture.molar_gibbs + self._property_modifiers['G']

    @material_property
    def _molar_volume_unmodified(self):
        return self.mixture.molar_volume

    @material_property
    def molar_volume(self):
        """
        Returns molar volume of the solid solution [m^3/mol]
        Aliased with self.V
        """
        return self.mixture.molar_volume + self._property_modifiers['dGdP']

    @material_property
    def molar_entropy(self):
        """
        Returns entropy of the solid solution [J]
        Aliased with self.S
        """
        return self.mixture.molar_entropy - self._property_modifiers['dGdT']

    @material_property
    def isothermal_bulk_modulus(self):
        """
        Returns isothermal bulk modulus of the solid solution [Pa]
        Aliased with self.K_T
        """
        K_T_orig = self.mixture.isothermal_bulk_modulus

        return self.molar_volume \
            / ((self._molar_volume_unmodified / K_T_orig) - self._property_modifiers['d2GdP2'])

    @material_property
    def shear_modulus(self):
        """
        Returns shear modulus of the solid solution [Pa]
        Aliased with self.G
        """
        return self.mixture.shear_modulus

    @material_property
    def thermal_expansivity(self):
        """
        Returns thermal expansion coefficient (alpha) of the solid solution [1/K]
        Aliased with self.alpha
        """
        return ((self.mixture.thermal_expansivity * self._molar_volume_unmodified)
                + self._property_modifiers['d2GdPdT']) / self.molar_volume

    @material_property
    def molar_heat_capacity_p(self):
        """
        Returns heat capacity at constant pressure of the solid solution [J/K/mol]
        Aliased with self.C_p
        """
        return self.mixture.molar_heat_capacity_p - self.temperature * self._property_modifiers['d2GdT2']


    """
    Properties from mineral parameters,
    Legendre transformations
    or Maxwell relations
    """

    @material_property
    def molar_mass(self):
        """
        Returns molar mass of the solid solution [kg/mol]
        """
        return self.mixture.molar_mass

    @material_property
    def formula(self):
        """
        Returns molar chemical formula of the solid solution
        """
        return self.mixture.formula

    @material_property
    def density(self):
        """
        Returns density of the solid solution [kg/m^3]
        Aliased with self.rho
        """
        return self.molar_mass / self.molar_volume

    @material_property
    def molar_internal_energy(self):
        """
        Returns molar internal energy of the mineral [J/mol]
        Aliased with self.energy
        """
        return self.molar_gibbs - self.pressure * self.molar_volume + self.temperature * self.molar_entropy

    @material_property
    def molar_helmholtz(self):
        """
        Returns molar Helmholtz free energy of the solid solution [J/mol]
        Aliased with self.helmholtz
        """
        return self.molar_gibbs - self.pressure * self.molar_volume


    @material_property
    def molar_enthalpy(self):
        """
        Returns molar enthalpy of the solid solution [J/mol]
        Aliased with self.H
        """
        return self.molar_gibbs + self.temperature * self.molar_entropy


    @material_property
    def adiabatic_bulk_modulus(self):
        """
        Returns adiabatic bulk modulus of the solid solution [Pa]
        Aliased with self.K_S
        """
        if self.temperature < 1.e-10:
            return self.isothermal_bulk_modulus
        else:
            return self.isothermal_bulk_modulus * self.molar_heat_capacity_p / self.molar_heat_capacity_v

    @material_property
    def isothermal_compressibility(self):
        """
        Returns isothermal compressibility of the solid solution (or inverse isothermal bulk modulus) [1/Pa]
        Aliased with self.K_T
        """
        return 1. / self.isothermal_bulk_modulus

    @material_property
    def adiabatic_compressibility(self):
        """
        Returns adiabatic compressibility of the solid solution (or inverse adiabatic bulk modulus) [1/Pa]
        Aliased with self.K_S
        """
        return 1. / self.adiabatic_bulk_modulus

    @material_property
    def p_wave_velocity(self):
        """
        Returns P wave speed of the solid solution [m/s]
        Aliased with self.v_p
        """
        return np.sqrt((self.adiabatic_bulk_modulus + 4. / 3. *
                        self.shear_modulus) / self.density)

    @material_property
    def bulk_sound_velocity(self):
        """
        Returns bulk sound speed of the solid solution [m/s]
        Aliased with self.v_phi
        """
        return np.sqrt(self.adiabatic_bulk_modulus / self.density)

    @material_property
    def shear_wave_velocity(self):
        """
        Returns shear wave speed of the solid solution [m/s]
        Aliased with self.v_s
        """
        return np.sqrt(self.shear_modulus / self.density)

    @material_property
    def grueneisen_parameter(self):
        """
        Returns grueneisen parameter of the solid solution [unitless]
        Aliased with self.gr
        """
        if self.temperature < 1.e-12:
            return 0.
        else:
            return self.thermal_expansivity * self.isothermal_bulk_modulus \
                * self.molar_volume / self.molar_heat_capacity_v

    @material_property
    def molar_heat_capacity_v(self):
        """
        Returns molar heat capacity at constant volume of the solid solution [J/K/mol]
        Aliased with self.C_v
        """
        return self.molar_heat_capacity_p - self.molar_volume * self.temperature \
            * self.thermal_expansivity * self.thermal_expansivity \
            * self.isothermal_bulk_modulus



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

# This module provides the functions required to process the
# standard burnman formula formats.
# tools.chemistry returns the number of atoms and molar mass of a compound
# given its unit formula as an argument.
# process_solution_chemistry returns information required to calculate
# solid solution properties from a set of endmember formulae

from __future__ import absolute_import
import re
import numpy as np
from scipy.linalg import lu
from scipy.optimize import fsolve
from fractions import Fraction
from collections import Counter
import pkgutil
from string import ascii_uppercase as ucase
from sympy import nsimplify

from .. import constants


def read_masses():
    """
    A simple function to read a file with a two column list of
    elements and their masses into a dictionary
    """
    datastream = pkgutil.get_data(
        'burnman', 'data/input_masses/atomic_masses.dat')
    datalines = [line.strip()
                 for line in datastream.decode('ascii').split('\n')
                 if line.strip()]
    lookup = dict()
    for line in datalines:
        data = "%".join(line.split("%")[:1]).split()
        if data != []:
            lookup[data[0]] = float(data[1])
    return lookup


"""
atomic_masses is a dictionary of atomic masses
"""
atomic_masses = read_masses()

"""
IUPAC_element_order provides a list of all the elements.
Element order is based loosely on electronegativity,
following the scheme suggested by IUPAC, except that H
comes after the Group 16 elements, not before them.
"""
IUPAC_element_order = ['v', 'Og', 'Rn', 'Xe', 'Kr', 'Ar', 'Ne', 'He',  # Group 18
                       'Fr', 'Cs', 'Rb', 'K', 'Na', 'Li',  # Group 1 (not H)
                       'Ra', 'Ba', 'Sr', 'Ca', 'Mg', 'Be',  # Group 2
                       'Lr', 'No', 'Md', 'Fm', 'Es', 'Cf', 'Bk', 'Cm',
                       'Am', 'Pu', 'Np', 'U', 'Pa', 'Th', 'Ac',  # Actinides
                       'Lu', 'Yb', 'Tm', 'Er', 'Ho', 'Dy', 'Tb', 'Gd',
                       'Eu', 'Sm', 'Pm', 'Nd', 'Pr', 'Ce', 'La',  # Lanthanides
                       'Y', 'Sc',  # Group 3
                       'Rf', 'Hf', 'Zr', 'Ti',  # Group 4
                       'Db', 'Ta', 'Nb', 'V',  # Group 5
                       'Sg', 'W', 'Mo', 'Cr',  # Group 6
                       'Bh', 'Re', 'Tc', 'Mn',  # Group 7
                       'Hs', 'Os', 'Ru', 'Fe',  # Group 8
                       'Mt', 'Ir', 'Rh', 'Co',  # Group 9
                       'Ds', 'Pt', 'Pd', 'Ni',  # Group 10
                       'Rg', 'Au', 'Ag', 'Cu',  # Group 11
                       'Cn', 'Hg', 'Cd', 'Zn',  # Group 12
                       'Nh', 'Tl', 'In', 'Ga', 'Al', 'B',  # Group 13
                       'Fl', 'Pb', 'Sn', 'Ge', 'Si', 'C',  # Group 14
                       'Mc', 'Bi', 'Sb', 'As', 'P', 'N',  # Group 15
                       'Lv', 'Po', 'Te', 'Se', 'S', 'O',  # Group 16
                       'H',  # hydrogen
                       'Ts', 'At', 'I', 'Br', 'Cl', 'F']  # Group 17


def dictionarize_formula(formula):
    """
    A function to read a chemical formula string and
    convert it into a dictionary

    Parameters
    ----------
    formula : string object
        Chemical formula, written in the XnYm format, where
        the formula has n atoms of element X and m atoms of element Y

    Returns
    -------
    f : dictionary object
        The same chemical formula, but expressed as a dictionary.
    """
    f = dict()
    elements = re.findall('[A-Z][^A-Z]*', formula)
    for element in elements:
        element_name = re.split('[0-9][^A-Z]*', element)[0]
        element_atoms = re.findall('[0-9][^A-Z]*', element)
        if len(element_atoms) == 0:
            element_atoms = Fraction(1.0)
        else:
            element_atoms = Fraction(element_atoms[0])
        f[element_name] = f.get(element_name, 0.0) + element_atoms

    return f


def sum_formulae(formulae, amounts=None):
    """
    Adds together a set of formulae.

    Parameters
    ----------
    formulae : list of dictionary or counter objects
        List of chemical formulae
    amounts : list of floats
        List of amounts of each formula

    Returns
    -------
    summed_formula : Counter object
        The sum of the user-provided formulae
    """
    if amounts is None:
        amounts = [1. for formula in formulae]
    else:
        assert (len(formulae) == len(amounts))

    summed_formula = Counter()
    for i, formula in enumerate(formulae):
        summed_formula.update(Counter({element: amounts[i] * n_atoms
                                       for (element, n_atoms)
                                       in formula.items()}))
    return summed_formula


def formula_mass(formula):
    """
    A function to take a chemical formula and compute the formula mass.

    Parameters
    ----------
    formula : dictionary or counter object
        A chemical formula

    Returns
    -------
    mass : float
        The mass per mole of formula
    """
    mass = sum(formula[element] * atomic_masses[element]
               for element in formula)
    return mass


def convert_formula(formula, to_type='mass', normalize=False):
    """
    Converts a chemical formula from one type (mass or molar)
    into the other. Renormalises amounts if normalize=True

    Parameters
    ----------
    formula : dictionary or counter object
        A chemical formula

    to_type : string, one of 'mass' or 'molar'
        Conversion type

    normalize : boolean
        Whether or not to normalize the converted formula to 1

    Returns
    -------
    f : dictionary
        The converted formula
    """

    if to_type == 'mass':
        f = {element: n_atoms * atomic_masses[element]
             for (element, n_atoms) in formula.items()}
    elif to_type == 'molar':
        f = {element: n_atoms / atomic_masses[element]
             for (element, n_atoms) in formula.items()}
    else:
        raise Exception('Value of parameter to_type not recognised. '
                        'Should be either "mass" or "molar".')

    if normalize:
        s = np.sum([n for (element, n) in f.items()])
        f = {element: n/s for (element, n) in f.items()}

    return f


def process_solution_chemistry(solution_model):
    """
    This function parses a class instance with a "formulas"
    attribute containing site information, e.g.

        [ '[Mg]3[Al]2Si3O12', '[Mg]3[Mg1/2Si1/2]2Si3O12' ]

    It outputs the bulk composition of each endmember
    (removing the site information), and also a set of
    variables and arrays which contain the site information.
    These are output in a format that can easily be used to
    calculate activities and gibbs free energies, given
    molar fractions of the phases and pressure
    and temperature where necessary.

    Parameters
    ----------
    solution_model : instance of class
        Class must have a "formulas" attribute, containing a
        list of chemical formulae with site information

    Returns
    -------
    none
        Nothing is returned from this function, but the solution_model
        object gains the following attributes.

    solution_formulae : list of dictionaries
        List of endmember formulae is output from site formula strings

    n_sites : integer
        Number of sites in the solid solution.
        Should be the same for all endmembers.

    sites : list of lists of strings
        A list of elements for each site in the solid solution

    site_names : list of strings
        A list of elements_site pairs in the solid solution, where
        each distinct site is given by a unique uppercase letter
        e.g. ['Mg_A', 'Fe_A', 'Al_A', 'Al_B', 'Si_B']

    n_occupancies : integer
        Sum of the number of possible elements on each of the sites
        in the solid solution.
        Example: A binary solution [[A][B],[B][C1/2D1/2]] would have
        n_occupancies = 5, with two possible elements on
        Site 1 and three on Site 2

    site_multiplicities : array of floats
        The number of each site per formula unit
        To simplify computations later, the multiplicities
        are repeated for each element on each site

    endmember_occupancies : 2d array of floats
        A 1D array for each endmember in the solid solution,
        containing the fraction of atoms of each element on each site.

    endmember_noccupancies : 2d array of floats
        A 1D array for each endmember in the solid solution,
        containing the number of atoms of each element on each site
        per mole of endmember.
    """
    formulae = solution_model.formulas
    n_sites = formulae[0].count('[')
    n_endmembers = len(formulae)

    # Check the number of sites is the same for all endmembers
    for i in range(n_endmembers):
        assert(formulae[i].count('[') == n_sites)

    solution_formulae = []
    sites = [[] for i in range(n_sites)]
    list_occupancies = []
    list_multiplicity = np.empty(shape=(n_sites))
    n_occupancies = 0

    # Number of unique site occupancies (e.g.. Mg on X etc.)
    for i_mbr in range(n_endmembers):
        solution_formula = dict()
        list_occupancies.append([[0] * len(sites[site])
                                for site in range(n_sites)])
        s = re.split(r'\[', formulae[i_mbr])[1:]

        for i_site, site_string in enumerate(s):
            site_split = re.split(r'\]', site_string)
            site_occupancy = site_split[0]

            mult = re.split('[A-Z][^A-Z]*', site_split[1])[0]
            if mult == '':
                list_multiplicity[i_site] = Fraction(1.0)
            else:
                list_multiplicity[i_site] = Fraction(mult)

            # Loop over elements on a site
            elements = re.findall('[A-Z][^A-Z]*', site_occupancy)

            for element in elements:
                # Find the element and proportion on the site
                element_split = re.split('([0-9][^A-Z]*)', element)
                element_on_site = element_split[0]
                if len(element_split) == 1:
                    proportion_element_on_site = Fraction(1.0)
                else:
                    proportion_element_on_site = Fraction(element_split[1])

                solution_formula[element_on_site] = solution_formula.get(
                    element_on_site, 0.0) + (list_multiplicity[i_site]
                                             * proportion_element_on_site)

                if element_on_site not in sites[i_site]:
                    n_occupancies += 1
                    sites[i_site].append(element_on_site)
                    i_el = sites[i_site].index(element_on_site)
                    for parsed_mbr in range(len(list_occupancies)):
                        list_occupancies[parsed_mbr][i_site].append(0)
                else:
                    i_el = sites[i_site].index(element_on_site)
                list_occupancies[i_mbr][i_site][i_el] = proportion_element_on_site

            # Loop over elements after site
            if len(site_split) != 1:
                not_in_site = str(filter(None, site_split[1]))
                not_in_site = not_in_site.replace(mult, '', 1)
                for enamenumber in re.findall('[A-Z][^A-Z]*', not_in_site):
                    element = list(
                        filter(None, re.split(r'(\d+)', enamenumber)))
                    # Look up number of atoms of element
                    if len(element) == 1:
                        nel = 1.
                    else:
                        nel = float(float(element[1]))
                    solution_formula[element[0]] = solution_formula.get(
                        element[0], 0.0) + nel

        solution_formulae.append(solution_formula)

    # Site occupancies and multiplicities
    endmember_occupancies = np.empty(shape=(n_endmembers, n_occupancies))
    site_multiplicities = np.empty(shape=(n_occupancies))
    for i_mbr in range(n_endmembers):
        n_element = 0
        for i_site in range(n_sites):
            for i_el in range(len(list_occupancies[i_mbr][i_site])):
                endmember_occupancies[i_mbr][
                    n_element] = list_occupancies[i_mbr][i_site][i_el]
                site_multiplicities[n_element] = list_multiplicity[i_site]
                n_element += 1

    # Site names
    solution_model.site_names = []
    for i, elements in enumerate(sites):
        for element in elements:
            solution_model.site_names.append('{0}_{1}'.format(element,
                                                              ucase[i]))

    # Finally, make attributes for solution model instance:
    solution_model.solution_formulae = solution_formulae
    solution_model.n_sites = n_sites
    solution_model.sites = sites
    solution_model.n_occupancies = n_occupancies
    solution_model.site_multiplicities = site_multiplicities
    solution_model.endmember_occupancies = endmember_occupancies
    solution_model.endmember_noccupancies = np.einsum('ij, j->ij',
                                                      endmember_occupancies,
                                                      site_multiplicities)


def site_occupancies_to_strings(site_species_names, site_multiplicities,
                                endmember_occupancies):
    """
    Converts a list of endmember site occupancies into a list
    of string representations of those occupancies.

    Parameters
    ----------
    site_species_names : 2D list of strings
        A list of list of strings, giving the names of the species
        which reside on each site.
        List of sites, each of which contains a list of the species
        occupying each site.

    site_multiplicities : numpy array of floats
        List of floats giving the multiplicity of each site
        Must be either the same length as the number of sites, or
        the same length as site_species_names
        (with an implied repetition of the same
        number for each species on a given site).

    endmember_occupancies : 2D numpy array of floats
        A list of site-species occupancies for each endmember.
        The first dimension loops over the endmembers, and the
        second dimension loops over the site-species occupancies for that endmember.
        The total number and order of occupancies must
        be the same as the strings in site_species_names.

    Returns
    -------
    site_formulae : list of strings
        A list of strings in standard burnman format.
        For example, [Mg]3[Al]2 would correspond to the
        classic two-site pyrope garnet.
    """

    # Site multiplicities should either be given on a per-site basis,
    # or a per-species basis
    if len(site_species_names) == len(site_multiplicities):
        site_mults = []

        for i, site in enumerate(site_species_names):
            for species in site:
                site_mults.append(site_multiplicities[i])

        site_multiplicities = site_mults

    elif len(endmember_occupancies[0]) != len(site_multiplicities):
        raise Exception(
            'Site multiplicities should either be given on a per-site basis or a per-species basis')

    site_formulae = []
    for mbr_occupancies in endmember_occupancies:
        i = 0
        site_formulae.append('')
        for site in site_species_names:
            amounts = mbr_occupancies[i:i+len(site)]
            mult = site_multiplicities[i]
            if np.abs(mult - 1.) < 1.e-12:
                mult = ''
            else:
                mult = str(nsimplify(mult))
            amounts /= sum(amounts)
            site_occupancy = formula_to_string(dict(zip(site, amounts)))
            site_formulae[-1] += '[{0}]{1}'.format(site_occupancy, mult)
            i += len(site)

    return site_formulae


def compositional_array(formulae):
    """
    Parameters
    ----------
    formulae : list of dictionaries
        List of chemical formulae

    Returns
    -------
    formula_array : 2D array of floats
        Array of endmember formulae

    elements : List of strings
        List of elements
    """
    elements = []
    for formula in formulae:
        for element in formula:
            if element not in elements:
                elements.append(element)

    formula_array = ordered_compositional_array(formulae, elements)

    return formula_array, elements


def ordered_compositional_array(formulae, elements):
    """
    Parameters
    ----------
    formulae : list of dictionaries
        List of chemical formulae

    elements : List of strings
        List of elements

    Returns
    -------
    formula_array : 2D array of floats
        Array of endmember formulae
    """
    formula_array = np.zeros(shape=(len(formulae), len(elements)))
    for idx, formula in enumerate(formulae):
        for element in formula:
            assert(element in elements)
            formula_array[idx][elements.index(element)] = formula[element]

    return formula_array


def formula_to_string(formula):
    """
    Parameters
    ----------
    formula : dictionary or counter
        Chemical formula

    Returns
    -------
    formula_string : string
        A formula string, with element order as given in the list
        IUPAC_element_order.
        If one or more keys in the dictionary are not one of the elements
        in the periodic table, then they are added at the end of the string.
    """

    formula_string = ''
    for e in IUPAC_element_order:
        if e in formula and np.abs(formula[e]) > 1.e-12:
            if np.abs(formula[e] - 1.) < 1.e-12:
                formula_string += e
            else:
                formula_string += e + str(nsimplify(formula[e]))

    for e in formula:
        if e not in IUPAC_element_order:
            if e in formula and np.abs(formula[e]) > 1.e-12:
                if np.abs(formula[e] - 1.) < 1.e-12:
                    formula_string += e
                else:
                    formula_string += e + str(nsimplify(formula[e]))

    return formula_string


def sort_element_list_to_IUPAC_order(element_list):
    """
    Parameters
    ----------
    element_list : list
        List of elements

    Returns
    -------
    sorted_list : list
        List of elements sorted into IUPAC order
    """
    sorted_list = [e for e in IUPAC_element_order if e in element_list]
    assert len(sorted_list) == len(element_list)
    return sorted_list


def convert_fractions(composite, phase_fractions, input_type, output_type):
    """
    Takes a composite with a set of user defined molar, volume
    or mass fractions (which do not have to be the fractions
    currently associated with the composite) and
    converts the fractions to molar, mass or volume.

    Conversions to and from mass require a molar mass to be
    defined for all phases. Conversions to and from volume
    require set_state to have been called for the composite.

    Parameters
    ----------
    composite : Composite
        Composite for which fractions are to be defined.

    phase_fractions : list of floats
        List of input phase fractions (of type input_type)

    input_type : string
        Input fraction type: 'molar', 'mass' or 'volume'

    output_type : string
        Output fraction type: 'molar', 'mass' or 'volume'

    Returns
    -------
    output_fractions : list of floats
        List of output phase fractions (of type output_type)
    """
    if input_type == 'volume' or output_type == 'volume':
        if composite.temperature is None:
            raise Exception(
                composite.to_string() + ".set_state(P, T) has not been called, so volume fractions are currently undefined. Exiting.")

    if input_type == 'molar':
        molar_fractions = phase_fractions
    if input_type == 'volume':
        total_moles = sum(
            volume_fraction / phase.molar_volume for volume_fraction,
            phase in zip(phase_fractions, composite.phases))
        molar_fractions = [volume_fraction / (phase.molar_volume * total_moles)
                           for volume_fraction, phase in zip(phase_fractions, composite.phases)]
    if input_type == 'mass':
        total_moles = sum(mass_fraction / phase.molar_mass for mass_fraction,
                          phase in zip(phase_fractions, composite.phases))
        molar_fractions = [mass_fraction / (phase.molar_mass * total_moles)
                           for mass_fraction, phase in zip(phase_fractions, composite.phases)]

    if output_type == 'volume':
        total_volume = sum(
            molar_fraction * phase.molar_volume for molar_fraction,
            phase in zip(molar_fractions, composite.phases))
        output_fractions = [molar_fraction * phase.molar_volume
                            / total_volume for molar_fraction, phase in zip(molar_fractions, composite.phases)]
    elif output_type == 'mass':
        total_mass = sum(molar_fraction * phase.molar_mass for molar_fraction,
                         phase in zip(molar_fractions, composite.phases))
        output_fractions = [molar_fraction * phase.molar_mass
                            / total_mass for molar_fraction, phase in zip(molar_fractions, composite.phases)]
    elif output_type == 'molar':
        output_fractions = molar_fractions

    return output_fractions


def chemical_potentials(assemblage, component_formulae):
    """
    The compositional space of the components does not have to be a
    superset of the compositional space of the assemblage. Nor do they have to
    compose an orthogonal basis.

    The components must each be described by a linear mineral combination

    The mineral compositions must be linearly independent

        Parameters
        ----------
        assemblage : list of classes
            List of material classes
            set_method and set_state should already have been used
            the composition of the solid solutions should also have been set

        component_formulae : list of dictionaries
            List of chemical component formula dictionaries
            No restriction on length

        Returns
        -------
        component_potentials : array of floats
            Array of chemical potentials of components

    """
    # Split solid solutions into their respective endmembers
    # Find the chemical potentials of all the endmembers
    from ..classes.solidsolution import SolidSolution
    endmember_list = []
    endmember_potentials = []
    for mineral in assemblage:
        if isinstance(mineral, SolidSolution):
            for member in mineral.endmembers:
                endmember_list.append(member[0])
            for potential in mineral.partial_gibbs:
                endmember_potentials.append(potential)
        else:
            endmember_list.append(mineral)
            endmember_potentials.append(mineral.gibbs)

    # Make an array of all the endmember formulae
    endmember_formulae = [endmember.params['formula']
                          for endmember in endmember_list]
    endmember_compositions, elements = compositional_array(endmember_formulae)

    pl, u = lu(endmember_compositions, permute_l=True)
    assert(min(np.dot(np.square(u), np.ones(shape=len(elements)))) > 1e-6), \
        'Endmember compositions do not form an independent set of basis vectors'

    # Make an array of component formulae with elements in the same order as
    # the endmember array
    component_compositions = ordered_compositional_array(
        component_formulae, elements)

    p = np.linalg.lstsq(endmember_compositions.T, component_compositions.T, rcond=-1)
    # rcond=-1 is old default to silence warning and be compatible with python 2.7
    for idx, error in enumerate(p[1]):
        assert (error < 1e-10), \
            'Component %d not defined by prescribed assemblage' % (idx + 1)

    # Create an array of endmember proportions which sum to each component
    # composition
    endmember_proportions = np.around(p[0], 10).T

    # Calculate the chemical potential of each component
    component_potentials = np.dot(endmember_proportions, endmember_potentials)
    return component_potentials


def fugacity(standard_material, assemblage):
    """
        Parameters
        ----------
        standard_material: class
            Material class
            set_method and set_state should already have been used
            material must have a formula as a dictionary parameter

        assemblage: list of classes
            List of material classes
            set_method and set_state should already have been used

        Returns
        -------
        fugacity : float
            Value of the fugacity of the component with respect to
            the standard material

    """
    component_formula = standard_material.params['formula']
    chemical_potential = chemical_potentials(
        assemblage, [component_formula])[0]

    fugacity = np.exp((chemical_potential - standard_material.gibbs) / (
        constants.gas_constant * assemblage[0].temperature))
    return fugacity


def relative_fugacity(standard_material, assemblage, reference_assemblage):
    """
        Parameters
        ----------
        standard_material: class
            Material class
            set_method and set_state should already have been used
            material must have a formula as a dictionary parameter

        assemblage: list of classes
            List of material classes
            set_method and set_state should already have been used

        reference_assemblage: list of classes
            List of material classes
            set_method and set_state should already have been used

        Returns
        -------
        relative_fugacity : float
            Value of the fugacity of the component in the assemblage
            with respect to the reference_assemblage

    """
    component_formula = standard_material.params['formula']
    chemical_potential = chemical_potentials(
        assemblage, [component_formula])[0]
    reference_chemical_potential = chemical_potentials(
        reference_assemblage, [component_formula])[0]

    relative_fugacity = np.exp((chemical_potential - reference_chemical_potential) / (
        constants.gas_constant * assemblage[0].temperature))
    return relative_fugacity


def equilibrium_pressure(minerals, stoichiometry, temperature,
                         pressure_initial_guess=1.e5):
    """
    Given a list of minerals, their reaction stoichiometries
    and a temperature of interest, compute the
    equilibrium pressure of the reaction.

    Parameters
    ----------
    minerals : list of minerals
        List of minerals involved in the reaction.

    stoichiometry : list of floats
        Reaction stoichiometry for the minerals provided.
        Reactants and products should have the opposite signs [mol]

    temperature : float
        Temperature of interest [K]

    pressure_initial_guess : optional float
        Initial pressure guess [Pa]

    Returns
    -------
    pressure : float
        The equilibrium pressure of the reaction [Pa]
    """
    def eqm(P, T):
        gibbs = 0.
        for i, mineral in enumerate(minerals):
            mineral.set_state(P[0], T)
            gibbs = gibbs + mineral.gibbs * stoichiometry[i]
        return gibbs

    pressure = fsolve(eqm, [pressure_initial_guess], args=(temperature))[0]

    return pressure


def equilibrium_temperature(minerals, stoichiometry, pressure, temperature_initial_guess=1000.):
    """
    Given a list of minerals, their reaction stoichiometries
    and a pressure of interest, compute the
    equilibrium temperature of the reaction.

    Parameters
    ----------
    minerals : list of minerals
        List of minerals involved in the reaction.

    stoichiometry : list of floats
        Reaction stoichiometry for the minerals provided.
        Reactants and products should have the opposite signs [mol]

    pressure : float
        Pressure of interest [Pa]

    temperature_initial_guess : optional float
        Initial temperature guess [K]

    Returns
    -------
    temperature : float
        The equilibrium temperature of the reaction [K]
    """
    def eqm(T, P):
        gibbs = 0.
        for i, mineral in enumerate(minerals):
            mineral.set_state(P, T[0])
            gibbs = gibbs + mineral.gibbs * stoichiometry[i]
        return gibbs

    temperature = fsolve(eqm, [temperature_initial_guess], args=(pressure))[0]

    return temperature


def invariant_point(minerals_r1, stoichiometry_r1,
                    minerals_r2, stoichiometry_r2,
                    pressure_temperature_initial_guess=[1.e9, 1000.]):
    """
    Given a list of minerals, their reaction stoichiometries
    and a pressure of interest, compute the
    equilibrium temperature of the reaction.

    Parameters
    ----------
    minerals : list of minerals
        List of minerals involved in the reaction.

    stoichiometry : list of floats
        Reaction stoichiometry for the minerals provided.
        Reactants and products should have the opposite signs [mol]

    pressure : float
        Pressure of interest [Pa]

    temperature_initial_guess : optional float
        Initial temperature guess [K]

    Returns
    -------
    temperature : float
        The equilibrium temperature of the reaction [K]
    """
    def eqm(PT):
        P, T = PT
        gibbs_r1 = 0.
        for i, mineral in enumerate(minerals_r1):
            mineral.set_state(P, T)
            gibbs_r1 = gibbs_r1 + mineral.gibbs * stoichiometry_r1[i]
        gibbs_r2 = 0.
        for i, mineral in enumerate(minerals_r2):
            mineral.set_state(P, T)
            gibbs_r2 = gibbs_r2 + mineral.gibbs * stoichiometry_r2[i]
        return [gibbs_r1, gibbs_r2]

    pressure, temperature = fsolve(eqm, pressure_temperature_initial_guess)
    return pressure, temperature


def hugoniot(mineral, P_ref, T_ref, pressures, reference_mineral=None):
    """
    Calculates the temperatures (and volumes) along a Hugoniot
    as a function of pressure according to the Hugoniot equation
    U2-U1 = 0.5*(p2 - p1)(V1 - V2) where U and V are the
    internal energies and volumes (mass or molar) and U = F + TS


    Parameters
    ----------
    mineral : mineral
        Mineral for which the Hugoniot is to be calculated.

    P_ref : float
        Reference pressure [Pa]

    T_ref : float
        Reference temperature [K]

    pressures : numpy array of floats
        Set of pressures [Pa] for which the Hugoniot temperature
        and volume should be calculated

    reference_mineral : mineral
        Mineral which is stable at the reference conditions
        Provides an alternative U_0 and V_0 when the reference
        mineral transforms to the mineral of interest at some
        (unspecified) pressure.

    Returns
    -------
    temperatures : numpy array of floats
        The Hugoniot temperatures at pressure

    volumes : numpy array of floats
        The Hugoniot volumes at pressure
    """

    def Ediff(T, mineral, P, P_ref, U_ref, V_ref):
        mineral.set_state(P, T[0])
        U = mineral.helmholtz + T[0] * mineral.S
        V = mineral.V

        return (U - U_ref) - 0.5 * (P - P_ref) * (V_ref - V)

    if reference_mineral is None:
        reference_mineral = mineral

    reference_mineral.set_state(P_ref, T_ref)
    U_ref = reference_mineral.helmholtz + T_ref * reference_mineral.S
    V_ref = reference_mineral.V

    temperatures = np.empty_like(pressures)
    volumes = np.empty_like(pressures)

    for i, P in enumerate(pressures):
        temperatures[i] = fsolve(
            Ediff, [T_ref], args=(mineral, P, P_ref, U_ref, V_ref))[0]
        volumes[i] = mineral.V

    return temperatures, volumes



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


# This is a standalone program that converts a tabulated version of the
# Stixrude and Lithgow-Bertelloni data format into the standard burnman
# format (printed to stdout)


import sys


def read_dataset(datafile):
    f = open(datafile, 'r')
    ds = []
    for line in f:
        ds.append(line.decode('utf-8').split())
    return ds

ds = read_dataset('HHPH2013_endmembers.dat')

print '# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences'
print '# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU GPL v2 or later.'
print ''
print ''
print '"""'
print 'HHPH_2013'
print 'Minerals from Holland et al 2013 and references therein'
print 'The values in this document are all in S.I. units,'
print 'unlike those in the original paper'
print 'File autogenerated using HHPHdata_to_burnman.py'
print '"""'
print ''
print 'from burnman.mineral import Mineral'
print 'from burnman.solidsolution import SolidSolution'
print 'from burnman.solutionmodel import *'
print 'from burnman.tools.chemistry import dictionarize_formula, formula_mass'
print ''

print '"""'
print 'ENDMEMBERS'
print '"""'
print ''

param_scales = [-1., -1.,  # not nubmers, so we won't scale
                1.e3, 1.e3,  # kJ -> J
                1.0,  # J/K/mol
                1.e-5,  # kJ/kbar/mol -> m^3/mol
                1.e3, 1.e-2, 1.e3, 1.e3,
                # kJ -> J and table conversion for b
                1.e-5,  # table conversion
                1.e8,  # kbar -> Pa
                1.0,  # no scale for K'0
                1.e-8]  # GPa -> Pa # no scale for eta_s


formula = '0'
for idx, m in enumerate(ds):
    if idx == 0:
        param_names = m
    else:
        print 'class', m[0].lower(), '(Mineral):'
        print '    def __init__(self):'
        print ''.join(['        formula=\'', m[1], '\''])
        print '        formula = dictionarize_formula(formula)'
        print '        self.params = {'
        print ''.join(['            \'name\': \'', m[0], '\','])
        print '            \'formula\': formula,'
        print '            \'equation_of_state\': \'hp_tmt\','
        for pid, param in enumerate(m):
            if pid > 1 and pid != 3 and pid < 6:
                print '            \'' + param_names[pid] + '\':', float(param) * param_scales[pid], ','

        print '            \'Cp\':', [round(float(m[i]) * param_scales[i], 10) for i in [6, 7, 8, 9]], ','
        for pid, param in enumerate(m):
            if pid > 9:
                print '            \'' + param_names[pid] + '\':', float(param) * param_scales[pid], ','

        print '            \'n\': sum(formula.values()),'
        print '            \'molar_mass\': formula_mass(formula)}'
        print ''
        print '        self.uncertainties = {'
        print '            \'' + param_names[3] + '\':', float(m[3]) * param_scales[3], '}'

        print '        Mineral.__init__(self)'
        print ''



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
SLB_2005
^^^^^^^^

Minerals from Stixrude & Lithgow-Bertelloni 2005 and references therein

"""
from __future__ import absolute_import
from ..classes import mineral_helpers as helpers
from ..classes.mineral import Mineral


class stishovite (Mineral):

    def __init__(self):
        self.params = {
            'formula': {'Si': 1., 'O': 2.},
            'equation_of_state': 'slb3',
            'V_0': 14.02e-6,
            'K_0': 314.0e9,
            'Kprime_0': 4.4,
            'G_0': 220.0e9,
            'Gprime_0': 1.6,
            'molar_mass': .0601,
            'n': 3,
            'Debye_0': 1044.,
            'grueneisen_0': 1.34,
            'q_0': 2.4,
            'eta_s_0': 5.0}

        Mineral.__init__(self)


class periclase (Mineral):

    def __init__(self):
        self.params = {
            'formula': {'Mg': 1., 'O': 1.},
            'equation_of_state': 'slb3',
            'V_0': 11.24e-6,
            'K_0': 161.0e9,
            'Kprime_0': 3.8,
            'G_0': 131.0e9,
            'Gprime_0': 2.1,
            'molar_mass': .0403,
            'n': 2,
            'Debye_0': 773.,
            'grueneisen_0': 1.5,
            'q_0': 1.5,
            'eta_s_0': 2.8}
        Mineral.__init__(self)


class wuestite (Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb3',
            'formula': {'Fe': 1., 'O': 1.},
            'V_0': 12.06e-6,
            'K_0': 152.0e9,
            'Kprime_0': 4.9,
            'G_0': 47.0e9,
            'Gprime_0': 0.7,
            'molar_mass': .0718,
            'n': 2,
            'Debye_0': 455.,
            'grueneisen_0': 1.28,
            'q_0': 1.5,
            'eta_s_0': 0.8}

        Mineral.__init__(self)


class mg_perovskite(Mineral):

    def __init__(self):
        self.params = {
            'formula': {'Mg': 1., 'Si': 1., 'O': 3.},
            'equation_of_state': 'slb3',
            'V_0': 24.45e-6,
            'K_0': 251.0e9,
            'Kprime_0': 4.1,
            'G_0': 175.0e9,
            'Gprime_0': 1.7,
            'molar_mass': .1000,
            'n': 5,
            'Debye_0': 1070.,
            'grueneisen_0': 1.48,
            'q_0': 1.4,
            'eta_s_0': 2.6}

        Mineral.__init__(self)


class fe_perovskite(Mineral):

    def __init__(self):
        self.params = {
            'formula': {'Fe': 1., 'Si': 1., 'O': 3.},
            'equation_of_state': 'slb3',
            'V_0': 25.48e-6,
            'K_0': 281.0e9,
            'Kprime_0': 4.1,
            'G_0': 138.0e9,
            'Gprime_0': 1.7,
            'molar_mass': .1319,
            'n': 5,
            'Debye_0': 841.,
            'grueneisen_0': 1.48,
            'q_0': 1.4,
            'eta_s_0': 2.1}

        Mineral.__init__(self)

mg_bridgmanite = mg_perovskite
fe_bridgmanite = fe_perovskite



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
import numpy as np

# Try to import the jit from numba.  If it is
# not available, just go with the standard
# python interpreter
try:
    import os
    if 'NUMBA_DISABLE_JIT' in os.environ and int(os.environ['NUMBA_DISABLE_JIT']) == 1:
        raise ImportError("NOOOO!")
    from numba import jit
except ImportError:
    def jit(fn):
        return fn

import scipy.integrate as integrate

from .. import constants

"""
Functions for the Debye model.  Note that this is not Mie-Grueneisen-Debye,
just Debye, so is pretty limited.  Combine this with Mie-Grueneisen and
Birch-Murnaghan to get a full EOS
"""


chebyshev_representation = np.array(
    [2.707737068327440945 / 2.0, 0.340068135211091751, -0.12945150184440869e-01,
     0.7963755380173816e-03, -
     0.546360009590824e-04, 0.39243019598805e-05,
     -0.2894032823539e-06, 0.217317613962e-07, -
     0.16542099950e-08,
     0.1272796189e-09, -
     0.987963460e-11, 0.7725074e-12, -
     0.607797e-13,
     0.48076e-14, -0.3820e-15, 0.305e-16, -0.24e-17])


@jit
def _chebval(x, c):
    """
    Evaluate a Chebyshev series at points x.
    This is just a lightly modified copy/paste job from the numpy
    implementation of the same function, copied over here to put a
    jit wrapper around it.
    """
    if len(c) == 1:
        c0 = c[0]
        c1 = 0
    elif len(c) == 2:
        c0 = c[0]
        c1 = c[1]
    else:
        x2 = 2 * x
        c0 = c[-2]
        c1 = c[-1]
        for i in range(3, len(c) + 1):
            tmp = c0
            c0 = c[-i] - c1
            c1 = tmp + c1 * x2
    return c0 + c1 * x


def debye_fn(x):
    """
    Evaluate the Debye function.  Takes the parameter
    xi = Debye_T/T
    """
    sol = integrate.quad(
        lambda xi: (xi * xi * xi) / (np.exp(xi) - 1.), 0.0, x)  # EQ B3
    return 3. * sol[0] / pow(x, 3.)


eps = np.finfo(float).eps
sqrt_eps = np.sqrt(np.finfo(float).eps)
log_eps = np.log(np.finfo(float).eps)


@jit
def debye_fn_cheb(x):
    """
    Evaluate the Debye function using a Chebyshev series expansion coupled with
    asymptotic solutions of the function.  Shamelessly adapted from the GSL implementation
    of the same function (Itself adapted from Collected Algorithms from ACM).
    Should give the same result as debye_fn(x) to near machine-precision.
    """
    val_infinity = 19.4818182068004875
    xcut = -log_eps

    assert(x > 0.0)  # check for invalid x

    if x < 2.0 * np.sqrt(2.0) * sqrt_eps:
        return 1.0 - 3.0 * x / 8.0 + x * x / 20.0
    elif x <= 4.0:
        t = x * x / 8.0 - 1.0
        c = _chebval(t, chebyshev_representation)
        return c - 0.375 * x
    elif x < -(np.log(2.0) + log_eps):
        nexp = int(np.floor(xcut / x))
        ex = np.exp(-x)
        xk = nexp * x
        rk = nexp
        sum = 0.0
        for i in range(nexp, 0, -1):
            xk_inv = 1.0 / xk
            sum *= ex
            sum += (((6.0 * xk_inv + 6.0) * xk_inv + 3.0) * xk_inv + 1.0) / rk
            rk -= 1.0
            xk -= x
        return val_infinity / (x * x * x) - 3.0 * sum * ex
    elif x < xcut:
        x3 = x * x * x
        sum = 6.0 + 6.0 * x + 3.0 * x * x + x3
        return (val_infinity - 3.0 * sum * np.exp(-x)) / x3
    else:
        return ((val_infinity / x) / x) / x


@jit
def thermal_energy(T, debye_T, n):
    """
    calculate the thermal energy of a substance.  Takes the temperature,
    the Debye temperature, and n, the number of atoms per molecule.
    Returns thermal energy in J/mol
    """
    if T <= eps:
        return 0.
    E_th = 3. * n * constants.gas_constant * T * debye_fn_cheb(debye_T / T)
    return E_th


@jit
def molar_heat_capacity_v(T, debye_T, n):
    """
    Heat capacity at constant volume.  In J/K/mol
    """
    if T <= eps:
        return 0.
    x = debye_T / T
    C_v = 3.0 * n * constants.gas_constant * \
        (4.0 * debye_fn_cheb(x) - 3.0 * x / (np.exp(x) - 1.0))
    return C_v


@jit
def helmholtz_free_energy(T, debye_T, n):
    """
    Helmholtz free energy of lattice vibrations in the Debye model.
    It is important to note that this does NOT include the zero
    point energy of vibration for the lattice.  As long as you are
    calculating relative differences in F, this should cancel anyways.
    In Joules.
    """
    if T <= eps:
        return 0.
    x = debye_T / T
    F = n * constants.gas_constant * T * \
        (3.0 * np.log(1.0 - np.exp(-x)) - debye_fn_cheb(x))
    return F


def entropy(T, debye_T, n):
    """
    Entropy due to lattice vibrations in the Debye model [J/K]
    """
    if T <= eps:
        return 0.
    x = debye_T / T
    S = n * constants.gas_constant * \
        (4. * debye_fn_cheb(x) - 3. * np.log(1.0 - np.exp(-x)))
    return S



from __future__ import absolute_import
# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2015 by the BurnMan team, released under the GNU
# GPL v2 or later.


import scipy.optimize as opt
from . import equation_of_state as eos
from ..tools.math import bracket
import warnings
import numpy as np

def bulk_modulus(volume, params):
    """
    Compute the bulk modulus as per the Morse potential
    equation of state.
    Returns bulk modulus in the same units as
    the reference bulk modulus.
    Pressure must be in :math:`[Pa]`.
    """

    VoverV0 = volume / params['V_0']
    x = (params['Kprime_0']  - 1.)*(1. - np.power(VoverV0, 1./3.))
    K = params['K_0']*( ( 2./(params['Kprime_0']  - 1.) *
                          np.power(VoverV0, -2./3.) *
                          (np.exp(2.*x) - np.exp(x)) ) +
                        ( np.power(VoverV0, -1./3.) *
                          (2.*np.exp(2.*x) - np.exp(x)) ) )
    return K

def shear_modulus(volume, params):
    """
    Shear modulus not currently implemented for this equation of state
    """
    return 0.

def morse_potential(VoverV0, params):
    """
    Equation for the Morse Potential equation of state,
    returns pressure in the same units that are supplied
    for the reference bulk modulus (params['K_0'])
    """
    x = (params['Kprime_0']  - 1.)*(1. - np.power(VoverV0, 1./3.))
    return ( 3. * params['K_0'] / (params['Kprime_0']  - 1.) *
             np.power(VoverV0, -2./3.) *
             (np.exp(2.*x) - np.exp(x)) ) + params['P_0']

def volume(pressure, params):
    """
    Get the Morse Potential volume at a
    reference temperature for a given pressure :math:`[Pa]`.
    Returns molar volume in :math:`[m^3]`
    """
    func = lambda V: morse_potential(V / params['V_0'], params) - pressure
    try:
        sol = bracket(func, params['V_0'], 1.e-2 * params['V_0'])
    except:
        raise ValueError(
            'Cannot find a volume, perhaps you are outside of the range of validity for the equation of state?')
    return opt.brentq(func, sol[0], sol[1])



class Morse(eos.EquationOfState):

    """
    Class for the isothermal Morse Potential equation of state
    detailed in :cite:`Stacey1981`.
    This equation of state has no temperature dependence.
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume :math:`[m^3]` as a function of pressure :math:`[Pa]`.
        """
        return volume(pressure, params)

    def pressure(self, temperature, volume, params):
        return morse_potential(volume / params['V_0'], params)

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`K_T` :math:`[Pa]` as a function of pressure :math:`[Pa]`,
        temperature :math:`[K]` and volume :math:`[m^3]`.
        """
        return bulk_modulus(volume, params)

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus :math:`K_s` of the mineral. :math:`[Pa]`.
        """
        return bulk_modulus(volume, params)

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus :math:`G` of the mineral. :math:`[Pa]`
        """
        return shear_modulus(volume, params)

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy :math:`\mathcal{S}` of the mineral. :math:`[J/K/mol]`
        """
        return 0.

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy :math:`\mathcal{E}` of the mineral. :math:`[J/mol]`
        """

        x = (params['Kprime_0'] - 1)*(1 - np.power(volume/params['V_0'], 1./3.))
        intPdV = ( 9./2. * params['V_0'] * params['K_0'] /
                   np.power(params['Kprime_0'] - 1., 2.) *
                   (2.*np.exp(x) - np.exp(2.*x) - 1.) )

        return -intPdV + params['E_0']

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy :math:`\mathcal{G}` of the mineral. :math:`[J/mol]`
        """
        return self.molar_internal_energy(pressure, temperature, volume, params) + volume*pressure

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.e99

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.e99

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[1/K]`
        """
        return 0.

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[unitless]`
        """
        return 0.

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        if 'E_0' not in params:
            params['E_0'] = 0.
        if 'P_0' not in params:
            params['P_0'] = 0.

        # If G and Gprime are not included this is presumably deliberate,
        # as we can model density and bulk modulus just fine without them,
        # so just add them to the dictionary as nans
        if 'G_0' not in params:
            params['G_0'] = float('nan')
        if 'Gprime_0' not in params:
            params['Gprime_0'] = float('nan')

        # Check that all the required keys are in the dictionary
        expected_keys = ['V_0', 'K_0', 'Kprime_0', 'G_0', 'Gprime_0']
        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # Finally, check that the values are reasonable.
        if params['P_0'] < 0.:
            warnings.warn('Unusual value for P_0', stacklevel=2)
        if params['V_0'] < 1.e-7 or params['V_0'] > 1.e-3:
            warnings.warn('Unusual value for V_0', stacklevel=2)
        if params['K_0'] < 1.e9 or params['K_0'] > 1.e13:
            warnings.warn('Unusual value for K_0', stacklevel=2)
        if params['Kprime_0'] < 0. or params['Kprime_0'] > 10.:
            warnings.warn('Unusual value for Kprime_0', stacklevel=2)
        if params['G_0'] < 0.0 or params['G_0'] > 1.e13:
            warnings.warn('Unusual value for G_0', stacklevel=2)
        if params['Gprime_0'] < -5. or params['Gprime_0'] > 10.:
            warnings.warn('Unusual value for Gprime_0', stacklevel=2)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


# This is a standalone program that converts a tabulated version of the
# Stixrude and Lithgow-Bertelloni data format into the standard burnman
# format (printed to stdout)


import sys
import os.path

if os.path.isfile('stx11ver.dat') == False:
    print('This code requires the data file stx11ver.dat.')
    print(
        'This file is bundled with the Perple_X data files, which can be found here:')
    print('http://www.perplex.ethz.ch/perplex/ibm_and_mac_archives/')
    print('')
    print('Please download the file and place it in this directory.')
    exit()


def read_dataset(datafile):
    f = open(datafile, 'r')
    ds = []
    for line in f:
        ds.append(line.split())
    return ds


def read_dataset_utf8(datafile):
    f = open(datafile, 'r')
    ds = []
    for line in f:
        ds.append(line.decode('utf-8').split())
    return ds


def process_stixrude():

    perplex_ds = read_dataset('stx11ver.dat')
    alias_mapping = read_dataset('perplex_slb_2011_names.txt')

    alias = {}
    accurate_values = {}
    configurational_entropies = {}
    for line in alias_mapping:
        alias[line[0]] = line[2]

    process = 0  # flag to process data or not (to cut out first lines)
    n = 0  # line number for each record
    for line in perplex_ds:
        if len(line) != 0:
            if process == 1:
                if line[0] == 'end':
                    n = 0
                else:
                    n = n + 1
                if n == 1:
                    name = alias[line[0]]
                    data = []
                    configurational_entropy = 'None'
                elif n == 3:
                    # F0, (n, ignored), -V0
                    data.extend([float(line[2]), -1.0 * float(line[8])])
                elif n == 4:
                    # K_0, K_prime, Debye_0, grueneisen_0, q_0, eta_s0,
                    # [configurational_and_magnetic_entropy]
                    data.extend(
                        [float(line[2]), float(line[5]), float(line[8]),
                         float(line[11]), float(line[14]), float(line[17])])
                    if len(line) == 21:
                        configurational_entropy = float(line[20])
                elif n == 5:
                    # mu_S0, mu_S0_prime
                    data.extend([float(line[2]), float(line[5])])
                    # Sort into correct order for processing: F, V, K, K',
                    #                        Debye, gruen, q, G, Gprime, etaS0
                    accurate_values[
                        name] = [data[0] / 1.e3, data[1] * 10., data[2] / 1.e4, data[3],
                                 data[4], data[5], data[6], data[8] / 1.e4, data[9], data[7]]
                    configurational_entropies[name] = configurational_entropy

            if line[0] == 'end_components':
                process = 1

    return accurate_values, configurational_entropies


accurate_values, configurational_entropies = process_stixrude()


ds = read_dataset_utf8('slb_2011.txt')
landau = read_dataset_utf8('slb_2011_landau.txt')
landau_params = {}
for line in landau:
    phase = str(line[0]).lower()
    if phase != '#':
        landau_params[phase] = [
            float(line[1]), float(line[2]) * 1.e-6, float(line[3])]

print '# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences'
print '# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU GPL v2 or later.'
print ''
print ''
print '"""'
print 'SLB_2011'
print 'Minerals from Stixrude & Lithgow-Bertelloni 2011 and references therein'
print 'File autogenerated using SLBdata_to_burnman.py'
print '"""'
print ''
print 'from __future__ import absolute_import'
print ''
print 'from ..mineral import Mineral'
print 'from ..solidsolution import SolidSolution'
print 'from ..solutionmodel import *'
print 'from ..tools.chemistry import dictionarize_formula, formula_mass'
print ''

param_scales = [-1., -1.,  # not numbers, so we won't scale
                1.e3, 1.e3,  # KJ -> J
                1.e-6, 1.e-6,  # cm^3/mol -> m^3/mol
                1.e9, 1.e9,  # GPa -> Pa
                1.0, 1.0,  # no scale for K'
                1.0, 1.0,  # no scale for Debye
                1.0, 1.0,  # no scale for gruneisen
                1.0, 1.0,  # no scale for q
                1.e9, 1.e9,  # GPa -> Pa
                1.0, 1.0,  # no scale for G'
                1.0, 1.0]  # no scale for eta_s


solutionfile = 'slb_2011_solutions.txt'
with open(solutionfile, 'r') as fin:
    print fin.read()
fin.close()

print '"""'
print 'ENDMEMBERS'
print '"""'
print ''
formula = '0'
for idx, m in enumerate(ds):
    if idx == 0:
        param_names = m
    else:
        print 'class', m[0].lower(), '(Mineral):'
        print '    def __init__(self):'
        print ''.join(['        formula=\'', m[1], '\''])
        print '        formula = dictionarize_formula(formula)'
        print '        self.params = {'
        print ''.join(['            \'name\': \'', m[0], '\','])
        print '            \'formula\': formula,'
        print '            \'equation_of_state\': \'slb3\','
        for pid, param in enumerate(accurate_values[m[0].lower()]):
            print '            \'' + param_names[(pid + 1) * 2] + '\':', float(param) * param_scales[(pid + 1) * 2], ','
        print '            \'n\': sum(formula.values()),'
        print '            \'molar_mass\': formula_mass(formula)}'
        print ''
        if landau_params.has_key(m[0].lower()) or configurational_entropies[m[0].lower()] != 'None':
            print '        self.property_modifiers = [',
            if landau_params.has_key(m[0].lower()):
                print '[\'landau\', {\'Tc_0\':', landau_params[m[0].lower()][0], ', \'S_D\':', landau_params[m[0].lower()][2], ', \'V_D\':', landau_params[m[0].lower()][1], '}]',
            if configurational_entropies[m[0].lower()] != 'None':
                print '[\'linear\', {\'delta_E\':', 0., ', \'delta_S\':', configurational_entropies[m[0].lower()], ', \'delta_V\':', 0., '}]',
            print ']'
            print ''
        print '        self.uncertainties = {'
        for pid, param in enumerate(m):
            if pid > 1 and pid % 2 == 1 and pid < 21:
                print '            \'' + param_names[pid] + '\':', float(param) * param_scales[pid], ','
        pid = 21
        param = m[pid]
        print '            \'' + param_names[pid] + '\':', float(param) * param_scales[pid], '}'
        print '        Mineral.__init__(self)'
        print ''


aliasfile = 'slb_2011_aliases.txt'
with open(aliasfile, 'r') as fin:
    print fin.read()
fin.close()



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
SLB_2011_ZSB_2013
^^^^^^^^^^^^^^^^^

Minerals from Stixrude & Lithgow-Bertelloni 2011, Zhang, Stixrude & Brodholt 2013, and references therein.

"""
from __future__ import absolute_import

from ..classes import mineral_helpers as helpers
from ..classes.mineral import Mineral


class stishovite (Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb3',
            'V_0': 14.02e-6,
            'K_0': 314.0e9,
            'Kprime_0': 3.8,
            'G_0': 220.0e9,
            'Gprime_0': 1.9,
            'molar_mass': .0601,
            'n': 3,
            'Debye_0': 1108.,
            'grueneisen_0': 1.37,
            'q_0': 2.8,
            'eta_s_0': 4.6}

        self.uncertainties = {
            'err_K_0': 8.e9,
            'err_Kprime_0': 0.1,
            'err_G_0': 12.e9,
            'err_Gprime_0': 0.1,
            'err_Debye_0': 13.,
            'err_grueneisen_0': .17,
            'err_q_0': 2.2,
            'err_eta_s_0': 1.0
        }

        Mineral.__init__(self)


class periclase (Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb3',
            'V_0': 11.24e-6,
            'K_0': 161.0e9,
            'Kprime_0': 3.8,
            'G_0': 131.0e9,
            'Gprime_0': 2.1,
            'molar_mass': .0403,
            'n': 2,
            'Debye_0': 767.,
            'grueneisen_0': 1.36,
            'q_0': 1.7,  # 1.7
            'eta_s_0': 2.8}  # 2.8

        self.uncertainties = {
            'err_K_0': 3.e9,
            'err_Kprime_0': .2,
            'err_G_0': 1.0e9,
            'err_Gprime_0': .1,
            'err_Debye_0': 9.,
            'err_grueneisen_0': .05,
            'err_q_0': .2,
            'err_eta_s_0': .2}

        Mineral.__init__(self)


class wuestite (Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb3',
            'V_0': 12.26e-6,
            'K_0': 179.0e9,
            'Kprime_0': 4.9,
            'G_0': 59.0e9,
            'Gprime_0': 1.4,
            'molar_mass': .0718,
            'n': 2,
            'Debye_0': 454.,
            'grueneisen_0': 1.53,
            'q_0': 1.7,  # 1.7
            'eta_s_0': -0.1}

        self.uncertainties = {
            'err_K_0': 1.e9,
            'err_Kprime_0': .2,
            'err_G_0': 1.e9,
            'err_Gprime_0': .1,
            'err_Debye_0': 21.,
            'err_grueneisen_0': .13,
            'err_q_0': 1.0,
            'err_eta_s_0': 1.0}

        Mineral.__init__(self)


class mg_perovskite(Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb3',
            'V_0': 24.45e-6,
            'K_0': 250.5e9,
            'Kprime_0': 4.01,
            'G_0': 172.9e9,
            'Gprime_0': 1.74,
            'molar_mass': .1000,
            'n': 5,
            'Debye_0': 905.9,
            'grueneisen_0': 1.44,
            'q_0': 1.09,
            'eta_s_0': 2.13}  # 2.6

        self.uncertainties = {
            'err_K_0': 3.e9,
            'err_Kprime_0': 0.1,
            'err_G_0': 2.e9,
            'err_Gprime_0': 0.0,
            'err_Debye_0': 5.,
            'err_grueneisen_0': .05,
            'err_q_0': .3,
            'err_eta_s_0': .3}

        Mineral.__init__(self)


class fe_perovskite(Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb3',
            'V_0': 25.49e-6,
            'K_0': 272.0e9,
            'Kprime_0': 4.1,
            'G_0': 133.0e9,
            'Gprime_0': 1.4,
            'molar_mass': .1319,
            'n': 5,
            'Debye_0': 871.,
            'grueneisen_0': 1.57,
            'q_0': 1.1,
            'eta_s_0': 2.3}  # 2.3

        self.uncertainties = {
            'err_K_0': 40e9,
            'err_Kprime_0': 1.,
            'err_G_0': 40e9,
            'err_Gprime_0': 0.0,
            'err_Debye_0': 26.,
            'err_grueneisen_0': .3,
            'err_q_0': 1.0,
            'err_eta_s_0': 1.0}

        Mineral.__init__(self)

mg_bridgmanite = mg_perovskite
fe_bridgmanite = fe_perovskite



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
import numpy as np
import warnings


class AveragingScheme(object):

    """
    Base class defining an interface for determining average
    elastic properties of a rock.  Given a list of volume
    fractions for the different mineral phases in a rock,
    as well as their bulk and shear moduli, an averaging
    will give back a single scalar values for the averages.
    New averaging schemes should define the functions
    average_bulk_moduli and average_shear_moduli, as
    specified here.
    """

    def average_bulk_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli :math:`K` for a composite. This defines the interface
        for this method, and is not implemented in the base class.

        Parameters
        ----------
        volumes : list of floats
            List of the volume of each phase in the composite. :math:`[m^3]`
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite. :math:`[Pa]`
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite. :math:`[Pa]`

        Returns
        -------

        K : float
            The average bulk modulus :math:`K`. :math:`[Pa]`
        """
        raise NotImplementedError("")

    def average_shear_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the shear moduli :math:`G` for a composite.  This defines the interface
        for this method, and is not implemented in the base class.

        Parameters
        ----------
        volumes : list of floats
            List of the volume of each phase in the composite. :math:`[m^3]`
        bulk_moduli : list of floats
            List of bulk moduli of each phase in the composite. :math:`[Pa]`
        shear_moduli : list of floats
            List of shear moduli of each phase in the composite. :math:`[Pa]`

        Returns
        -------

        G : float
            The average shear modulus :math:`G`. :math:`[Pa]`
        """
        raise NotImplementedError("")

    def average_density(self, volumes, densities):
        """
        Average the densities of a composite, given a list of volume
        fractions and densitites. This is implemented in the base class,
        as how to calculate it is not dependent on the geometry of the rock.
        The formula for density is given by

        .. math::
            \\rho = \\frac{\\Sigma_i \\rho_i V_i }{\\Sigma_i V_i}

        Parameters
        ----------
        volumes : list of floats
            List of the volume of each phase in the composite. :math:`[m^3]`
        densities : list of floats
            List of densities of each phase in the composite. :math:`[kg/m^3]`

        Returns
        -------
        rho : float
           Density :math:`\\rho`. :math:`[kg/m^3]`
        """
        total_mass = np.sum(np.array(densities) * np.array(volumes))
        total_vol = np.sum(np.array(volumes))  # should sum to one
        density = total_mass / total_vol
        return density

    def average_thermal_expansivity(self, volumes, alphas):
        """
        thermal expansion coefficient of the mineral :math:`\\alpha`. :math:`[1/K]`
        """
        total_vol = np.sum(np.array(volumes))
        return np.sum(np.array(alphas) * np.array(volumes)) / total_vol

    def average_heat_capacity_v(self, fractions, c_v):
        # TODO: double-check that the formula we use is appropriate here.
        """
        Averages the heat capacities at constant volume :math:`C_V` by molar fractions
        as in eqn. (16) in :cite:`Ita1992`.

        Parameters
        ----------

        fractions : list of floats
            List of molar fractions of each phase in the composite (should sum to 1.0).
        c_v : list of floats
            List of heat capacities at constant volume :math:`C_V` of each phase in the composite. :math:`[J/K/mol]`

        Returns
        -------

        c_v : float
          heat capacity at constant volume of the composite :math:`C_V`. :math:`[J/K/mol]`
        """
        return np.sum(np.array(fractions) * np.array(c_v))

    def average_heat_capacity_p(self, fractions, c_p):
        # TODO: double-check that the formula we use is correct.
        """
        Averages the heat capacities at constant pressure :math:`C_P` by molar fractions.

        Parameters
        ----------
        fractions : list of floats
            List of molar fractions of each phase in the composite (should sum to 1.0).
        c_p : list of floats
            List of heat capacities at constant pressure :math:`C_P` of each phase in the composite. :math:`[J/K/mol]`

        Returns
        -------
        c_p : float
          heat capacity at constant pressure :math:`C_P` of the composite. :math:`[J/K/mol]`
        """
        return np.sum(np.array(fractions) * np.array(c_p))


class VoigtReussHill(AveragingScheme):

    """
    Class for computing the Voigt-Reuss-Hill average for elastic properties.
    This derives from :class:`burnman.averaging_schemes.averaging_scheme`, and implements
    the :func:`burnman.averaging_schemes.averaging_scheme.average_bulk_moduli` and
    :func:`burnman.averaging_schemes.averaging_scheme.average_shear_moduli` functions.
    """

    def average_bulk_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli of a composite with the Voigt-Reuss-Hill average, given by:

        .. math::
            K_{VRH} = \\frac{K_V + K_R}{2}

        This is simply a shorthand for an arithmetic average of the bounds given
        by :class:`burnman.averaging_schemes.voigt` and :class:`burnman.averaging_schemes.reuss`.

        Parameters
        ----------
        volumes : list of floats
            List of the volume of each phase in the composite. :math:`[m^3]`
        bulk_moduli : list of floats
            List of bulk moduli :math:`K` of each phase in the composite. :math:`[Pa]`
        shear_moduli : list of floats
            List of shear moduli :math:`G` of each phase in the composite.
            Not used in this average. :math:`[Pa]`

        Returns
        -------

        K : float
            The Voigt-Reuss-Hill average bulk modulus :math:`K_{VRH}`. :math:`[Pa]`
        """
        return voigt_reuss_hill_function(volumes, bulk_moduli)

    def average_shear_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the shear moduli :math:`G` of a composite with the Voigt-Reuss-Hill average, given by:

        .. math::
            G_{VRH} = \\frac{G_V + G_R}{2}

        This is simply a shorthand for an arithmetic average of the bounds given
        by :class:`burnman.averaging_schemes.voigt` and :class:`burnman.averaging_schemes.reuss`.

        Parameters
        ----------
        volumes : list of floats
            List of the volume of each phase in the composite :math:`[m^3]`
        bulk_moduli : list of floats
            List of bulk moduli :math:`K` of each phase in the composite
            Not used in this average. :math:`[Pa]`
        shear_moduli : list of floats
            List of shear moduli :math:`G` of each phase in the composite :math:`[Pa]`

        Returns
        -------

        G : float
            The Voigt-Reuss-Hill average shear modulus :math:`G_{VRH}`. :math:`[Pa]`
        """
        return voigt_reuss_hill_function(volumes, shear_moduli)


class Voigt(AveragingScheme):

    """
    Class for computing the Voigt (iso-strain) bound for elastic properties.
    This derives from :class:`burnman.averaging_schemes.averaging_scheme`, and implements
    the :func:`burnman.averaging_schemes.averaging_scheme.average_bulk_moduli` and
    :func:`burnman.averaging_schemes.averaging_scheme.average_shear_moduli` functions.
    """

    def average_bulk_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli of a composite :math:`K` with the Voigt (iso-strain)
        bound, given by:

        .. math::
            K_V = \\Sigma_i V_i K_i

        Parameters
        ----------
        volumes : list of floats
            List of the volume of each phase in the composite. :math:`[m^3]`
        bulk_moduli : list of floats
            List of bulk moduli :math:`K` of each phase in the composite. :math:`[Pa]`
        shear_moduli : list of floats
            List of shear moduli :math:`G` of each phase in the composite.
            Not used in this average. :math:`[Pa]`

        Returns
        -------

        K : float
            The Voigt average bulk modulus :math:`K_V`. :math:`[Pa]`
        """
        return voigt_average_function(volumes, bulk_moduli)

    def average_shear_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the shear moduli of a composite with the Voigt (iso-strain)
        bound, given by:

        .. math::
            G_V = \\Sigma_i V_i G_i

        Parameters
        ----------
        volumes : list of floats
            List of the volume of each phase in the composite. :math:`[m^3]`
        bulk_moduli : list of floats
            List of bulk moduli :math:`K` of each phase in the composite.
            Not used in this average. :math:`[Pa]`
        shear_moduli : list of floats
            List of shear moduli :math:`G` of each phase in the composite. :math:`[Pa]`

        Returns
        -------

        G : float
            The Voigt average shear modulus :math:`G_V`. :math:`[Pa]`
        """
        return voigt_average_function(volumes, shear_moduli)


class Reuss(AveragingScheme):

    """
    Class for computing the Reuss (iso-stress) bound for elastic properties.
    This derives from :class:`burnman.averaging_schemes.averaging_scheme`, and implements
    the :func:`burnman.averaging_schemes.averaging_scheme.average_bulk_moduli` and
    :func:`burnman.averaging_schemes.averaging_scheme.average_shear_moduli` functions.
    """

    def average_bulk_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli of a composite with the Reuss (iso-stress)
        bound, given by:

        .. math::
            K_R = \\left(\\Sigma_i \\frac{V_i}{K_i} \\right)^{-1}

        Parameters
        ----------
        volumes : list of floats
            List of the volume of each phase in the composite. :math:`[m^3]`
        bulk_moduli : list of floats
            List of bulk moduli :math:`K` of each phase in the composite. :math:`[Pa]`
        shear_moduli : list of floats
            List of shear moduli :math:`G` of each phase in the composite.
            Not used in this average. :math:`[Pa]`

        Returns
        -------

        K : float
            The Reuss average bulk modulus :math:`K_R`. :math:`[Pa]`
        """
        return reuss_average_function(volumes, bulk_moduli)

    def average_shear_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the shear moduli of a composite with the Reuss (iso-stress)
        bound, given by:

        .. math::
            G_R = \\left( \\Sigma_i \\frac{V_i}{G_i} \\right)^{-1}

        Parameters
        ----------
        volumes : list of floats
            List of the volume of each phase in the composite. :math:`[m^3]`
        bulk_moduli : list of floats
            List of bulk moduli :math:`K` of each phase in the composite.
            Not used in this average. :math:`[Pa]`
        shear_moduli : list of floats
            List of shear moduli :math:`G` of each phase in the composite. :math:`[Pa]`

        Returns
        -------

        G : float
            The Reuss average shear modulus :math:`G_R`. :math:`[Pa]`
        """
        return reuss_average_function(volumes, shear_moduli)


class HashinShtrikmanUpper(AveragingScheme):

    """
    Class for computing the upper Hashin-Shtrikman bound for elastic properties.
    This derives from :class:`burnman.averaging_schemes.averaging_scheme`, and implements
    the :func:`burnman.averaging_schemes.averaging_scheme.average_bulk_moduli`
    and :func:`burnman.averaging_schemes.averaging_scheme.average_shear_moduli` functions.
    Implements formulas from :cite:`Watt1976`.  The Hashin-Shtrikman bounds
    are tighter than the Voigt and Reuss bounds because they make the
    additional assumption that the orientation of the phases are statistically
    isotropic.  In some cases this may be a good assumption, and in others it
    may not be.
    """

    def average_bulk_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli of a composite with the upper Hashin-Shtrikman bound.
        Implements Formulas from :cite:`Watt1976`, which are too lengthy to reproduce here.

        Parameters
        ----------
        volumes : list of floats
            List of the volume of each phase in the composite. :math:`[m^3]`
        bulk_moduli : list of floats
            List of bulk moduli :math:`K` of each phase in the composite. :math:`[Pa]`
        shear_moduli : list of floats
            List of shear moduli :math:`G` of each phase in the composite. :math:`[Pa]`

        Returns
        -------

        K : float
            The upper Hashin-Shtrikman average bulk modulus :math:`K`. :math:`[Pa]`
        """

        K_n = max(bulk_moduli)
        G_n = max(shear_moduli)

        vol_frac = volumes / sum(volumes)

        alpha_n = -3. / (3. * K_n + 4. * G_n)
        A_n = 0
        for i in range(len(vol_frac)):
            if bulk_moduli[i] != K_n:
                A_n += vol_frac[i] / (1. / (bulk_moduli[i] - K_n) - alpha_n)

        K_upper = K_n + A_n / (1. + alpha_n * A_n)
        return K_upper

    def average_shear_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the shear moduli of a composite with the upper Hashin-Shtrikman bound.
        Implements Formulas from :cite:`Watt1976`, which are too lengthy to reproduce here.

        Parameters
        ----------
        volumes : list of floats
            List of the volume of each phase in the composite. :math:`[m^3]`
        bulk_moduli : list of floats
            List of bulk moduli :math:`K` of each phase in the composite. :math:`[Pa]`
        shear_moduli : list of floats
            List of shear moduli :math:`G` of each phase in the composite. :math:`[Pa]`

        Returns
        -------

        G : float
            The upper Hashin-Shtrikman average shear modulus :math:`G`. :math:`[Pa]`
        """

        K_n = max(bulk_moduli)
        G_n = max(shear_moduli)

        vol_frac = volumes / sum(volumes)

        beta_n = -3. * (K_n + 2. * G_n) / (5. * G_n * (3. * K_n + 4. * G_n))
        B_n = 0
        for i in range(len(vol_frac)):
            if shear_moduli[i] != G_n:
                B_n += vol_frac[i] / (
                    1. / (2. * (shear_moduli[i] - G_n)) - beta_n)

        G_upper = G_n + (0.5) * B_n / (1. + beta_n * B_n)
        return G_upper


class HashinShtrikmanLower(AveragingScheme):

    """
    Class for computing the lower Hashin-Shtrikman bound for elastic properties.
    This derives from :class:`burnman.averaging_schemes.averaging_scheme`, and implements
    the :func:`burnman.averaging_schemes.averaging_scheme.average_bulk_moduli`
    and :func:`burnman.averaging_schemes.averaging_scheme.average_shear_moduli` functions.
    Implements Formulas from :cite:`Watt1976`.  The Hashin-Shtrikman bounds
    are tighter than the Voigt and Reuss bounds because they make the
    additional assumption that the orientation of the phases are statistically
    isotropic.  In some cases this may be a good assumption, and in others it
    may not be.
    """

    def average_bulk_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli of a composite with the lower Hashin-Shtrikman bound.
        Implements Formulas from :cite:`Watt1976`, which are too lengthy to reproduce here.

        Parameters
        ----------
        volumes : list of floats
            List of the volume of each phase in the composite. :math:`[m^3]`
        bulk_moduli : list of floats
            List of bulk moduli :math:`K` of each phase in the composite. :math:`[Pa]`
        shear_moduli : list of floats
            List of shear moduli :math:`G` of each phase in the composite. :math:`[Pa]`

        Returns
        -------

        K : float
            The lower Hashin-Shtrikman average bulk modulus :math:`K`. :math:`[Pa]`
        """

        K_1 = min(bulk_moduli)
        G_1 = min(shear_moduli)

        vol_frac = volumes / sum(volumes)

        alpha_1 = -3. / (3. * K_1 + 4. * G_1)
        A_1 = 0
        for i in range(len(vol_frac)):
            if bulk_moduli[i] != K_1:
                A_1 += vol_frac[i] / (1. / (bulk_moduli[i] - K_1) - alpha_1)

        K_lower = K_1 + A_1 / (1. + alpha_1 * A_1)
        return K_lower

    def average_shear_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the shear moduli of a composite with the lower Hashin-Shtrikman bound.
        Implements Formulas from :cite:`Watt1976`, which are too lengthy to reproduce here.

        Parameters
        ----------
        volumes : list of floats
            List of volumes of each phase in the composite. :math:`[m^3]`.
        bulk_moduli : list of floats
            List of bulk moduli :math:`K` of each phase in the composite. :math:`[Pa]`.
        shear_moduli : list of floats
            List of shear moduli :math:`G` of each phase in the composite. :math:`[Pa]`

        Returns
        -------

        G : float
            The lower Hashin-Shtrikman average shear modulus :math:`G`. :math:`[Pa]`
        """

        K_1 = min(bulk_moduli)
        G_1 = min(shear_moduli)

        vol_frac = volumes / sum(volumes)

        beta_1 = -3. * (K_1 + 2. * G_1) / (5. * G_1 * (3. * K_1 + 4. * G_1))
        B_1 = 0
        for i in range(len(vol_frac)):
            if shear_moduli[i] != G_1:
                B_1 += vol_frac[i] / (
                    1. / (2. * (shear_moduli[i] - G_1)) - beta_1)

        G_lower = G_1 + (0.5) * B_1 / (1. + beta_1 * B_1)
        return G_lower


class HashinShtrikmanAverage(AveragingScheme):

    """
    Class for computing arithmetic mean of the Hashin-Shtrikman bounds on elastic properties.
    This derives from :class:`burnman.averaging_schemes.averaging_scheme`, and implements
    the :func:`burnman.averaging_schemes.averaging_scheme.average_bulk_moduli`
    and :func:`burnman.averaging_schemes.averaging_scheme.average_shear_moduli` functions.
    """

    def __init__(self):
        self.upper = HashinShtrikmanUpper()
        self.lower = HashinShtrikmanLower()

    def average_bulk_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli of a composite with the arithmetic mean of the upper
        and lower Hashin-Shtrikman bounds.

        Parameters
        ----------
        volumes : list of floats
            List of the volumes of each phase in the composite. :math:`[m^3]`
        bulk_moduli : list of floats
            List of bulk moduli :math:`K` of each phase in the composite. :math:`[Pa]`
        shear_moduli : list of floats
            List of shear moduli :math:`G` of each phase in the composite.
            Not used in this average. :math:`[Pa]`

        Returns
        -------

        K : float
            The arithmetic mean of the Hashin-Shtrikman bounds on bulk modulus :math:`K`. :math:`[Pa]`
        """
        return (self.upper.average_bulk_moduli(volumes, bulk_moduli, shear_moduli)
                + self.lower.average_bulk_moduli(volumes, bulk_moduli, shear_moduli)) / 2.0

    def average_shear_moduli(self, volumes, bulk_moduli, shear_moduli):
        """
        Average the bulk moduli of a composite with the arithmetic mean of the upper
        and lower Hashin-Shtrikman bounds.

        Parameters
        ----------
        volumes : list of floats
            List of the volumes of each phase in the composite. [m^3].
        bulk_moduli : list of floats
            List of bulk moduli :math:`K` of each phase in the composite.
            Not used in this average. :math:`[Pa]`
        shear_moduli : list of floats
            List of shear moduli :math:`G` of each phase in the composite. :math:`[Pa]`

        Returns
        -------

        G : float
            The arithmetic mean of the Hashin-Shtrikman bounds on shear modulus :math:`G`. :math:`[Pa]`
        """
        return (self.upper.average_shear_moduli(volumes, bulk_moduli, shear_moduli)
                + self.lower.average_shear_moduli(volumes, bulk_moduli, shear_moduli)) / 2.0


def voigt_average_function(phase_volume, X):
    """
    Do Voigt (iso-strain) average.  Rather like
    resistors in series.  Called by voigt and
    voigt_reuss_hill classes, takes a list of
    volumes and moduli, returns a modulus.
    """
    vol_frac = phase_volume / np.sum(phase_volume)
    X_voigt = sum(f * x for f, x in zip(vol_frac, X))
    return X_voigt


def reuss_average_function(phase_volume, X):
    """
    Do Reuss (iso-stress) average.  Rather like
    resistors in parallel.  Called by reuss and
    voigt_reuss_hill classes, takes a list of
    volumes and moduli, returns a modulus.
    """
    vol_frac = phase_volume / np.sum(phase_volume)
    for f, x in zip(vol_frac, X):
        if x <= 0 and np.abs(f) > np.finfo(float).eps:
            warnings.warn("Oops, called reuss_average with Xi<=0!")
            return 0.0
    X_reuss = 1. / sum(f / x for f, x in zip(vol_frac, X))
    return X_reuss


def voigt_reuss_hill_function(phase_volume, X):
    """
    Do Voigt-Reuss-Hill average (arithmetic mean
    of Voigt and Reuss bounds).  Called by
    voigt_reuss_hill class, takes a list of
    volumes and moduli, returns a modulus.
    """
    X_vrh = (voigt_average_function(phase_volume, X)
             + reuss_average_function(phase_volume, X)) / 2.0
    return X_vrh



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
from __future__ import print_function

import numpy as np

from ..tools.math import unit_normalize
from .material import Material, material_property

try:  # numpy.block was new in numpy version 1.13.0.
    block = np.block([[np.ones((3, 3)), 2.*np.ones((3, 3))],
                      [2.*np.ones((3, 3)), 4.*np.ones((3, 3))]])
except:
    block = np.array(np.bmat([[[[1.]*3]*3, [[2.]*3]*3],
                              [[[2.]*3]*3, [[4.]*3]*3]] ))
voigt_compliance_factors = block


class AnisotropicMaterial(Material):
    """
    A base class for anisotropic elastic materials. The base class
    is initialised with a density and a full isentropic stiffness tensor
    in Voigt notation. It can then be interrogated to find the values of
    different properties, such as bounds on seismic velocities.
    There are also several functions which can be called to calculate
    properties along directions oriented with respect to the isentropic
    elastic tensor.

    See :cite:`Mainprice2011`
    and https://docs.materialsproject.org/methodology/elasticity/
    for mathematical descriptions of each function.
    """

    def __init__(self, rho, cijs):

        self._isentropic_stiffness_tensor = cijs
        self._rho = rho

        assert cijs.shape == (6, 6), 'cijs must be in Voigt notation (6x6)'
        assert np.allclose(cijs.T, cijs), 'stiffness_tensor must be symmetric'

        Material.__init__(self)

    def _voigt_index_to_ij(self, m):
        """
        Returns the ij (or kl) indices of the
        stiffness tensor which correspond to those
        of the Voigt notation m (or n).
        """
        if m == 3:
            return 1, 2
        elif m == 4:
            return 0, 2
        elif m == 5:
            return 0, 1
        else:
            return m, m

    def _voigt_notation_to_stiffness_tensor(self, voigt_notation):
        """
        Converts a stiffness tensor in Voigt notation (6x6 matrix)
        to the full fourth rank tensor (3x3x3x3 matrix).
        """
        stiffness_tensor = np.zeros([3, 3, 3, 3])
        for m in range(6):
            i, j = self._voigt_index_to_ij(m)
            for n in range(6):
                k, l = self._voigt_index_to_ij(n)
                stiffness_tensor[i][j][k][l] = voigt_notation[m][n]
                stiffness_tensor[j][i][k][l] = voigt_notation[m][n]
                stiffness_tensor[i][j][l][k] = voigt_notation[m][n]
                stiffness_tensor[j][i][l][k] = voigt_notation[m][n]
        return stiffness_tensor

    def _voigt_notation_to_compliance_tensor(self, voigt_notation):
        return self._voigt_notation_to_stiffness_tensor(np.divide(voigt_notation,
                                                                  voigt_compliance_factors))

    @material_property
    def isentropic_stiffness_tensor(self):
        return self._isentropic_stiffness_tensor

    @material_property
    def full_isentropic_stiffness_tensor(self):
        return self._voigt_notation_to_stiffness_tensor(self.isentropic_stiffness_tensor)

    @material_property
    def isentropic_compliance_tensor(self):
        return np.linalg.inv(self.isentropic_stiffness_tensor)

    @material_property
    def full_isentropic_compliance_tensor(self):
        return self._voigt_notation_to_compliance_tensor(self.isentropic_compliance_tensor)

    @material_property
    def density(self):
        return self._rho

    @material_property
    def isentropic_bulk_modulus_voigt(self):
        """
        Computes the isentropic bulk modulus (Voigt bound)
        """
        K = np.sum([[self.isentropic_stiffness_tensor[i][k]
                     for k in range(3)]
                    for i in range(3)])/9.
        return K

    @material_property
    def isentropic_bulk_modulus_reuss(self):
        """
        Computes the isentropic bulk modulus (Reuss bound)
        """
        beta = np.sum([[self.isentropic_compliance_tensor[i][k] for k in range(3)] for i in range(3)])
        return 1./beta

    @material_property
    def isentropic_bulk_modulus_vrh(self):
        """
        Computes the isentropic bulk modulus (Voigt-Reuss-Hill average)
        """
        return 0.5*(self.isentropic_bulk_modulus_voigt + self.isentropic_bulk_modulus_reuss)

    @material_property
    def isentropic_shear_modulus_voigt(self):
        """
        Computes the isentropic shear modulus (Voigt bound)
        """
        G = ( np.sum([self.isentropic_stiffness_tensor[i][i] for i in [0, 1, 2]]) +
              np.sum([self.isentropic_stiffness_tensor[i][i] for i in [3, 4, 5]])*3. -
              ( self.isentropic_stiffness_tensor[0][1] +
                self.isentropic_stiffness_tensor[1][2] +
                self.isentropic_stiffness_tensor[2][0] )) / 15.
        return G

    @material_property
    def isentropic_shear_modulus_reuss(self):
        """
        Computes the isentropic shear modulus (Reuss bound)
        """
        beta =  ( np.sum([self.isentropic_compliance_tensor[i][i] for i in [0, 1, 2]])*4. +
                  np.sum([self.isentropic_compliance_tensor[i][i] for i in [3, 4, 5]])*3. -
                  ( self.isentropic_compliance_tensor[0][1] +
                    self.isentropic_compliance_tensor[1][2] +
                    self.isentropic_compliance_tensor[2][0])*4. ) / 15.
        return 1./beta

    @material_property
    def isentropic_shear_modulus_vrh(self):
        """
        Computes the shear modulus (Voigt-Reuss-Hill average)
        """
        return 0.5*(self.isentropic_shear_modulus_voigt
                    + self.isentropic_shear_modulus_reuss)

    @material_property
    def isentropic_universal_elastic_anisotropy(self):
        """
        Compute the universal elastic anisotropy
        """
        return ( 5.*(self.isentropic_shear_modulus_voigt/self.isentropic_shear_modulus_reuss) +
                 (self.isentropic_bulk_modulus_voigt/self.isentropic_bulk_modulus_reuss) - 6. )

    @material_property
    def isentropic_isotropic_poisson_ratio(self):
        """
        Compute mu, the isotropic Poisson ratio
        (a description of the laterial response to loading)
        """
        return ((3.*self.isentropic_bulk_modulus_vrh
                 - 2.*self.isentropic_shear_modulus_vrh)
                / (6.*self.isentropic_bulk_modulus_vrh
                   + 2.*self.isentropic_shear_modulus_vrh) )

    def christoffel_tensor(self, propagation_direction):
        """
        Computes the Christoffel tensor from an elastic stiffness
        tensor and a propagation direction for a seismic wave
        relative to the stiffness tensor

        T_ik = C_ijkl n_j n_l
        """
        propagation_direction = unit_normalize(propagation_direction)
        Tik = np.tensordot(np.tensordot(self.full_isentropic_stiffness_tensor,
                                        propagation_direction,
                                        axes=([1],[0])),
                           propagation_direction,
                           axes=([2],[0]))
        return Tik

    def isentropic_linear_compressibility(self, direction):
        """
        Computes the linear isentropic compressibility in a given direction
        relative to the stiffness tensor
        """
        direction = unit_normalize(direction)
        Sijkk = np.einsum('ijkk', self.full_isentropic_compliance_tensor)
        beta = Sijkk.dot(direction).dot(direction)
        return beta

    def isentropic_youngs_modulus(self, direction):
        """
        Computes the isentropic Youngs modulus in a given direction
        relative to the stiffness tensor
        """
        direction = unit_normalize(direction)
        Sijkl = self.full_isentropic_compliance_tensor
        S = Sijkl.dot(direction).dot(direction).dot(direction).dot(direction)
        return 1./S

    def isentropic_shear_modulus(self, plane_normal, shear_direction):
        """
        Computes the isentropic shear modulus on a plane in a given
        shear direction relative to the stiffness tensor
        """
        plane_normal = unit_normalize(plane_normal)
        shear_direction = unit_normalize(shear_direction)

        assert np.abs(plane_normal.dot(shear_direction)) < np.finfo(float).eps, 'plane_normal and shear_direction must be orthogonal'
        Sijkl = self.full_isentropic_compliance_tensor
        G = Sijkl.dot(shear_direction).dot(plane_normal).dot(shear_direction).dot(plane_normal)
        return 0.25/G

    def isentropic_poissons_ratio(self,
                                  axial_direction,
                                  lateral_direction):
        """
        Computes the isentropic poisson ratio given loading and response
        directions relative to the stiffness tensor
        """

        axial_direction = unit_normalize(axial_direction)
        lateral_direction = unit_normalize(lateral_direction)
        assert np.abs(axial_direction.dot(lateral_direction)) < np.finfo(float).eps, 'axial_direction and lateral_direction must be orthogonal'

        Sijkl = self.full_isentropic_compliance_tensor
        x = axial_direction
        y = lateral_direction
        nu = -(Sijkl.dot(y).dot(y).dot(x).dot(x) /
               Sijkl.dot(x).dot(x).dot(x).dot(x) )
        return nu

    def wave_velocities(self, propagation_direction):
        """
        Computes the compressional wave velocity, and two
        shear wave velocities in a given propagation direction

        Returns two lists, containing the wave speeds and
        directions of particle motion relative to the stiffness tensor
        """
        propagation_direction = unit_normalize(propagation_direction)

        Tik = self.christoffel_tensor(propagation_direction)

        eigenvalues, eigenvectors = np.linalg.eig(Tik)

        idx = eigenvalues.argsort()[::-1]
        eigenvalues = np.real(eigenvalues[idx])
        eigenvectors = eigenvectors[:,idx]
        velocities = np.sqrt(eigenvalues/self.density)

        return velocities, eigenvectors

def voigt_array_from_cijs(cijs, index_lists):
    """
    Takes a list of cijs and a list of list of tuples corresponding to
    the positions of each cij in the Voigt form matrix.
    Note that the indices run from 0--5, not 1--6.
    """
    C = np.zeros([6, 6])
    for i, index_list in enumerate(index_lists):
        for indices in index_list:
            C[indices] = cijs[i]
            C[indices[::-1]] = cijs[i]
    return C

class IsotropicMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [C12, C44] (i.e. lambda and mu, the Lame parameters)
    """
    def __init__(self, rho, cijs):

        assert len(cijs) == 2
        cijs = list(cijs)
        cijs.insert(0, cijs[0] + 2.*cijs[1]) # C11 = C12 + 2C44
        index_lists = [[(0, 0), (1, 1), (2, 2)], # C11
                       [(0, 1), (0, 2), (1, 2)], # C12
                       [(3, 3), (4, 4), (5, 5)]] # C44

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))

class CubicMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [C11, C12, C44]
    """
    def __init__(self, rho, cijs):

        assert len(cijs) == 3
        index_lists = [[(0, 0), (1, 1), (2, 2)], # C11
                       [(0, 1), (0, 2), (1, 2)], # C12
                       [(3, 3), (4, 4), (5, 5)]] # C44

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))

class HexagonalMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [C11, C12, C13, C33, C44]
    """
    def __init__(self, rho, cijs):
        assert len(cijs) == 5
        cijs = list(cijs)
        cijs.append((cijs[0] - cijs[1])/2.) # C66 = (C11-C12)/2.

        index_lists = [[(0, 0), (1, 1)], # C11
                       [(0, 1)], # C12
                       [(0, 2), (1, 2)], # C13
                       [(2, 2)], # C33
                       [(3, 3), (4, 4)], # C44
                       [(5, 5)]] # C66

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))

class TetragonalMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [C11, C12, C13, C33, C44, C66] or
    [C11, C12, C13, C16, C33, C44, C66]
    """
    def __init__(self, rho, cijs):
        if len(cijs) == 6:
            # Tetragonal I / Laue class 4/mmm
            index_lists = [[(0, 0), (1, 1)], # C11
                           [(0, 1)], # C12
                           [(0, 2), (1, 2)], # C13
                           [(2, 2)], # C33
                           [(3, 3), (4, 4)], # C44
                           [(5, 5)]] # C66
        elif len(cijs) == 7:
            # Tetragonal II / Laue class 4/m
            cijs = list(cijs)
            cijs.insert(4, -cijs[3]) # C26 = -C16
            index_lists = [[(0, 0), (1, 1)], # C11
                           [(0, 1)], # C12
                           [(0, 2), (1, 2)], # C13
                           [(0, 5)], # C16
                           [(1, 5)], # C26
                           [(2, 2)], # C33
                           [(3, 3), (4, 4)], # C44
                           [(5, 5)]] # C66
        else:
            raise Exception('Tetragonal materials should have '
                            'either 6 or 7 independent Cijs')

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))


class RhombohedralMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [C11, C12, C13, C14, C33, C44, C66] or
    [C11, C12, C13, C14, C15, C33, C44, C66]
    """
    def __init__(self, rho, cijs):
        cijs = list(cijs)
        if len(cijs) == 7:
            # Rhombohedral I / Laue class \bar{3}m
            cijs.insert(4, -cijs[3]) # C24 = -C14
            index_lists = [[(0, 0), (1, 1)], # C11
                           [(0, 1)], # C12
                           [(0, 2), (1, 2)], # C13
                           [(0, 3), (4, 5)], # C14
                           [(1, 3)], # C24
                           [(2, 2)], # C33
                           [(3, 3), (4, 4)], # C44
                           [(5, 5)]] # C66

        elif len(cijs) == 8:
            # Rhombohedral II / Laue class \bar{3}
            cijs.insert(4, -cijs[3]) # C24 = -C14
            cijs.insert(6, -cijs[5]) # C25 = -C15
            index_lists = [[(0, 0), (1, 1)], # C11
                           [(0, 1)], # C12
                           [(0, 2), (1, 2)], # C13
                           [(0, 3), (4, 5)], # C14
                           [(1, 3)], # C24
                           [(0, 4)], # C15
                           [(1, 4), (3, 5)], # C25
                           [(2, 2)], # C33
                           [(3, 3), (4, 4)], # C44
                           [(5, 5)]] # C66

        else:
            raise Exception('Rhombohedral materials should have '
                            'either 7 or 8 independent Cijs')

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))

class OrthorhombicMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [C11, C12, C13, C22, C23, C33, C44, C55, C66]
    """
    def __init__(self, rho, cijs):

        assert len(cijs) == 9
        index_lists = [[(0, 0)], # C11
                       [(0, 1)], # C12
                       [(0, 2)], # C13
                       [(1, 1)], # C22
                       [(1, 2)], # C23
                       [(2, 2)], # C33
                       [(3, 3)], # C44
                       [(4, 4)], # C55
                       [(5, 5)]] # C66

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))


class MonoclinicMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [C11, C12, C13, C15, C22, C23, C25, C33, C35, C44, C46, C55, C66]
    """
    def __init__(self, rho, cijs):

        assert len(cijs) == 13
        index_lists = [[(0, 0)], # C11
                       [(0, 1)], # C12
                       [(0, 2)], # C13
                       [(0, 4)], # C15
                       [(1, 1)], # C22
                       [(1, 2)], # C23
                       [(1, 4)], # C25
                       [(2, 2)], # C33
                       [(2, 4)], # C35
                       [(3, 3)], # C44
                       [(3, 5)], # C46
                       [(4, 4)], # C55
                       [(5, 5)]] # C66

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))

class TriclinicMaterial(AnisotropicMaterial):
    """
    A class derived from the AnisotropicMaterial base class
    Initialization takes two input parameters; rho and
    [Cij, where 1<=i<=6 and i<=j<=6]
    """
    def __init__(self, rho, cijs):

        assert len(cijs) == 21
        index_lists=[[(i, j)] for i in range(6) for j in range(i, 6)]

        AnisotropicMaterial.__init__(
            self, rho, voigt_array_from_cijs(cijs, index_lists))
